import collections
import fractions
import sys

import pytest
from packaging.specifiers import SpecifierSet

import asdf
from asdf import AsdfFile, config_context
from asdf.exceptions import AsdfManifestURIMismatchWarning, AsdfSerializationError, ValidationError
from asdf.extension import (
    Compressor,
    Converter,
    ConverterProxy,
    Extension,
    ExtensionManager,
    ExtensionProxy,
    ManifestExtension,
    TagDefinition,
    Validator,
    get_cached_extension_manager,
)
from asdf.extension._manager import _resolve_type
from asdf.testing.helpers import roundtrip_object


class MinimumExtension:
    extension_uri = "asdf://somewhere.org/extensions/minimum-1.0"


class MinimumExtensionSubclassed(Extension):
    extension_uri = "asdf://somewhere.org/extensions/minimum-1.0"


class FullExtension:
    extension_uri = "asdf://somewhere.org/extensions/full-1.0"

    def __init__(
        self,
        converters=None,
        compressors=None,
        validators=None,
        asdf_standard_requirement=None,
        tags=None,
        legacy_class_names=None,
    ):
        self._converters = [] if converters is None else converters
        self._compressors = [] if compressors is None else compressors
        self._validators = [] if validators is None else validators
        self._asdf_standard_requirement = asdf_standard_requirement
        self._tags = [] if tags is None else tags
        self._legacy_class_names = [] if legacy_class_names is None else legacy_class_names

    @property
    def converters(self):
        return self._converters

    @property
    def compressors(self):
        return self._compressors

    @property
    def validators(self):
        return self._validators

    @property
    def asdf_standard_requirement(self):
        return self._asdf_standard_requirement

    @property
    def tags(self):
        return self._tags

    @property
    def legacy_class_names(self):
        return self._legacy_class_names


class MinimumConverter:
    def __init__(self, tags=None, types=None):
        if tags is None:
            self._tags = []
        else:
            self._tags = tags

        if types is None:
            self._types = []
        else:
            self._types = types

    @property
    def tags(self):
        return self._tags

    @property
    def types(self):
        return self._types

    def to_yaml_tree(self, obj, tag, ctx):
        return "to_yaml_tree result"

    def from_yaml_tree(self, obj, tag, ctx):
        return "from_yaml_tree result"


class FullConverter(MinimumConverter):
    def select_tag(self, obj, tags, ctx):
        return "select_tag result"


class MinimalCompressor(Compressor):
    @staticmethod
    def compress(data):
        return b""

    @property
    def label(self):
        return b"mini"


class MinimalValidator(Validator):
    schema_property = "fail"
    tags = ["**"]

    def validate(self, fail, node, schema):
        if fail:
            yield ValidationError("Node was doomed to fail")


# Some dummy types for testing converters:
class FooType:
    pass


class SubFooType(FooType):
    pass


class BarType:
    pass


class BazType:
    pass


def test_extension_proxy_maybe_wrap():
    extension = MinimumExtension()
    proxy = ExtensionProxy.maybe_wrap(extension)
    assert proxy.delegate is extension
    assert ExtensionProxy.maybe_wrap(proxy) is proxy

    with pytest.raises(TypeError, match=r"Extension must implement the Extension interface"):
        ExtensionProxy.maybe_wrap(object())


def test_extension_proxy():
    # Test with minimum properties:
    extension = MinimumExtension()
    proxy = ExtensionProxy(extension)

    assert isinstance(proxy, Extension)

    assert proxy.extension_uri == "asdf://somewhere.org/extensions/minimum-1.0"
    assert proxy.legacy_class_names == set()
    assert proxy.asdf_standard_requirement == SpecifierSet()
    assert proxy.converters == []
    assert proxy.compressors == []
    assert proxy.validators == []
    assert proxy.tags == []
    assert proxy.types == []
    assert proxy.tag_mapping == []
    assert proxy.url_mapping == []
    assert proxy.delegate is extension
    assert proxy.legacy is False
    assert proxy.package_name is None
    assert proxy.package_version is None
    assert proxy.class_name == "asdf._tests.test_extension.MinimumExtension"

    # The subclassed version should have the same defaults:
    extension = MinimumExtensionSubclassed()
    subclassed_proxy = ExtensionProxy(extension)
    assert subclassed_proxy.extension_uri == proxy.extension_uri
    assert subclassed_proxy.legacy_class_names == proxy.legacy_class_names
    assert subclassed_proxy.asdf_standard_requirement == proxy.asdf_standard_requirement
    assert subclassed_proxy.converters == proxy.converters
    assert subclassed_proxy.compressors == proxy.compressors
    assert subclassed_proxy.validators == proxy.validators
    assert subclassed_proxy.tags == proxy.tags
    assert subclassed_proxy.types == proxy.types
    assert subclassed_proxy.tag_mapping == proxy.tag_mapping
    assert subclassed_proxy.url_mapping == proxy.url_mapping
    assert subclassed_proxy.delegate is extension
    assert subclassed_proxy.legacy == proxy.legacy
    assert subclassed_proxy.package_name == proxy.package_name
    assert subclassed_proxy.package_version == proxy.package_name
    assert subclassed_proxy.class_name == "asdf._tests.test_extension.MinimumExtensionSubclassed"

    # Test with all properties present:
    converters = [MinimumConverter(tags=["asdf://somewhere.org/extensions/full/tags/foo-*"], types=[])]
    compressors = [MinimalCompressor()]
    validators = [MinimalValidator()]
    extension = FullExtension(
        converters=converters,
        compressors=compressors,
        validators=validators,
        asdf_standard_requirement=">=1.4.0",
        tags=["asdf://somewhere.org/extensions/full/tags/foo-1.0"],
        legacy_class_names=["foo.extensions.SomeOldExtensionClass"],
    )
    proxy = ExtensionProxy(extension, package_name="foo", package_version="1.2.3")

    assert proxy.extension_uri == "asdf://somewhere.org/extensions/full-1.0"
    assert proxy.legacy_class_names == {"foo.extensions.SomeOldExtensionClass"}
    assert proxy.asdf_standard_requirement == SpecifierSet(">=1.4.0")
    assert proxy.converters == [ConverterProxy(c, proxy) for c in converters]
    assert proxy.compressors == compressors
    assert proxy.validators == validators
    assert len(proxy.tags) == 1
    assert proxy.tags[0].tag_uri == "asdf://somewhere.org/extensions/full/tags/foo-1.0"
    assert proxy.types == []
    assert proxy.tag_mapping == []
    assert proxy.url_mapping == []
    assert proxy.delegate is extension
    assert proxy.legacy is False
    assert proxy.package_name == "foo"
    assert proxy.package_version == "1.2.3"
    assert proxy.class_name == "asdf._tests.test_extension.FullExtension"

    # Should fail when the input is not one of the two extension interfaces:
    with pytest.raises(TypeError, match=r"Extension must implement the Extension interface"):
        ExtensionProxy(object)

    # Should fail with a bad converter:
    with pytest.raises(TypeError, match=r"Converter must implement the .* interface"):
        ExtensionProxy(FullExtension(converters=[object()]))

    # Should fail with a bad compressor:
    with pytest.raises(TypeError, match=r"Extension property 'compressors' must contain instances of .*"):
        ExtensionProxy(FullExtension(compressors=[object()]))

    # Should fail with a bad validator
    with pytest.raises(TypeError, match=r"Extension property 'validators' must contain instances of .*"):
        ExtensionProxy(FullExtension(validators=[object()]))

    # Unparsable ASDF Standard requirement:
    with pytest.raises(ValueError, match=r"Invalid specifier:.*"):
        ExtensionProxy(FullExtension(asdf_standard_requirement="asdf-standard >= 1.4.0"))

    # Unrecognized ASDF Standard requirement type:
    with pytest.raises(TypeError, match=r"Extension property 'asdf_standard_requirement' must be str or None"):
        ExtensionProxy(FullExtension(asdf_standard_requirement=object()))

    # Bad tag:
    with pytest.raises(TypeError, match=r"Extension property 'tags' must contain str or .* values"):
        ExtensionProxy(FullExtension(tags=[object()]))

    # Bad legacy class names:
    with pytest.raises(TypeError, match=r"Extension property 'legacy_class_names' must contain str values"):
        ExtensionProxy(FullExtension(legacy_class_names=[object]))


def test_extension_proxy_tags():
    """
    The tags behavior is a tad complex, so they get their own test.
    """
    foo_tag_uri = "asdf://somewhere.org/extensions/full/tags/foo-1.0"
    foo_tag_def = TagDefinition(
        foo_tag_uri,
        schema_uris="asdf://somewhere.org/extensions/full/schemas/foo-1.0",
        title="Some tag title",
        description="Some tag description",
    )

    bar_tag_uri = "asdf://somewhere.org/extensions/full/tags/bar-1.0"
    bar_tag_def = TagDefinition(
        bar_tag_uri,
        schema_uris="asdf://somewhere.org/extensions/full/schemas/bar-1.0",
        title="Some other tag title",
        description="Some other tag description",
    )

    # The converter should return only the tags
    # supported by the extension.
    converter = FullConverter(tags=["**"])
    extension = FullExtension(tags=[foo_tag_def], converters=[converter])
    proxy = ExtensionProxy(extension)
    assert proxy.converters[0].tags == [foo_tag_uri]

    # The converter should not return tags that
    # its patterns do not match.
    converter = FullConverter(tags=["**/foo-1.0"])
    extension = FullExtension(tags=[foo_tag_def, bar_tag_def], converters=[converter])
    proxy = ExtensionProxy(extension)
    assert proxy.converters[0].tags == [foo_tag_uri]

    # The process should still work if the extension property
    # contains str instead of TagDescription.
    converter = FullConverter(tags=["**/foo-1.0"])
    extension = FullExtension(tags=[foo_tag_uri, bar_tag_uri], converters=[converter])
    proxy = ExtensionProxy(extension)
    assert proxy.converters[0].tags == [foo_tag_uri]


def test_extension_proxy_hash_and_eq():
    extension = MinimumExtension()
    proxy1 = ExtensionProxy(extension)
    proxy2 = ExtensionProxy(extension, package_name="foo", package_version="1.2.3")

    assert proxy1 == proxy2
    assert hash(proxy1) == hash(proxy2)
    assert proxy1 != extension
    assert proxy2 != extension


def test_extension_proxy_repr():
    proxy = ExtensionProxy(MinimumExtension(), package_name="foo", package_version="1.2.3")
    assert "class: asdf._tests.test_extension.MinimumExtension" in repr(proxy)
    assert "package: foo==1.2.3" in repr(proxy)
    assert "legacy: False" in repr(proxy)

    proxy = ExtensionProxy(MinimumExtension())
    assert "class: asdf._tests.test_extension.MinimumExtension" in repr(proxy)
    assert "package: (none)" in repr(proxy)
    assert "legacy: False" in repr(proxy)


def test_extension_manager():
    converter1 = FullConverter(
        tags=[
            "asdf://somewhere.org/extensions/full/tags/foo-*",
            "asdf://somewhere.org/extensions/full/tags/bar-*",
        ],
        types=[
            FooType,
            "asdf._tests.test_extension.BarType",
        ],
    )
    converter2 = FullConverter(
        tags=[
            "asdf://somewhere.org/extensions/full/tags/baz-*",
        ],
        types=[BazType],
    )
    converter3 = FullConverter(
        tags=[
            "asdf://somewhere.org/extensions/full/tags/foo-*",
        ],
        types=[
            FooType,
            BarType,
        ],
    )
    extension1 = FullExtension(
        converters=[converter1, converter2],
        tags=[
            "asdf://somewhere.org/extensions/full/tags/foo-1.0",
            "asdf://somewhere.org/extensions/full/tags/baz-1.0",
        ],
    )
    extension2 = FullExtension(
        converters=[converter3],
        tags=[
            "asdf://somewhere.org/extensions/full/tags/foo-1.0",
        ],
    )

    manager = ExtensionManager([extension1, extension2])

    assert manager.extensions == [ExtensionProxy(extension1), ExtensionProxy(extension2)]

    assert manager.handles_tag("asdf://somewhere.org/extensions/full/tags/foo-1.0") is True
    assert manager.handles_tag("asdf://somewhere.org/extensions/full/tags/bar-1.0") is False
    assert manager.handles_tag("asdf://somewhere.org/extensions/full/tags/baz-1.0") is True

    assert manager.handles_type(FooType) is True
    assert manager.handles_type(SubFooType) is False
    # This should return True even though BarType was listed
    # as string class name:
    assert manager.handles_type(BarType) is True
    assert manager.handles_type(BazType) is True

    assert (
        manager.get_tag_definition("asdf://somewhere.org/extensions/full/tags/foo-1.0").tag_uri
        == "asdf://somewhere.org/extensions/full/tags/foo-1.0"
    )
    assert (
        manager.get_tag_definition("asdf://somewhere.org/extensions/full/tags/baz-1.0").tag_uri
        == "asdf://somewhere.org/extensions/full/tags/baz-1.0"
    )
    with pytest.raises(KeyError, match=r"No support available for YAML tag.*"):
        manager.get_tag_definition("asdf://somewhere.org/extensions/full/tags/bar-1.0")

    assert manager.get_converter_for_tag("asdf://somewhere.org/extensions/full/tags/foo-1.0").delegate is converter1
    assert manager.get_converter_for_tag("asdf://somewhere.org/extensions/full/tags/baz-1.0").delegate is converter2
    with pytest.raises(KeyError, match=r"No support available for YAML tag.*"):
        manager.get_converter_for_tag("asdf://somewhere.org/extensions/full/tags/bar-1.0")

    assert manager.get_converter_for_type(FooType).delegate is converter1
    assert manager.get_converter_for_type(BarType).delegate is converter1
    assert manager.get_converter_for_type(BazType).delegate is converter2
    with pytest.raises(KeyError, match=r"\"No support available for Python type .*\""):
        manager.get_converter_for_type(object)
    with pytest.raises(KeyError, match=r"\"No support available for Python type .*\""):
        manager.get_converter_for_type(SubFooType)


def test_get_cached_extension_manager():
    extension = MinimumExtension()
    extension_manager = get_cached_extension_manager([extension])
    assert get_cached_extension_manager([extension]) is extension_manager
    assert get_cached_extension_manager([MinimumExtension()]) is not extension_manager


def test_tag_definition():
    tag_def = TagDefinition(
        "asdf://somewhere.org/extensions/foo/tags/foo-1.0",
        schema_uris="asdf://somewhere.org/extensions/foo/schemas/foo-1.0",
        title="Some title",
        description="Some description",
    )

    assert tag_def.tag_uri == "asdf://somewhere.org/extensions/foo/tags/foo-1.0"
    assert tag_def.schema_uris == ["asdf://somewhere.org/extensions/foo/schemas/foo-1.0"]
    assert tag_def.title == "Some title"
    assert tag_def.description == "Some description"

    assert "URI: asdf://somewhere.org/extensions/foo/tags/foo-1.0" in repr(tag_def)

    tag_def = TagDefinition(
        "asdf://somewhere.org/extensions/foo/tags/foo-1.0",
        schema_uris=[
            "asdf://somewhere.org/extensions/foo/schemas/foo-1.0",
            "asdf://somewhere.org/extensions/foo/schemas/base-1.0",
        ],
        title="Some title",
        description="Some description",
    )

    assert tag_def.schema_uris == [
        "asdf://somewhere.org/extensions/foo/schemas/foo-1.0",
        "asdf://somewhere.org/extensions/foo/schemas/base-1.0",
    ]

    with pytest.raises(ValueError, match=r"URI patterns are not permitted in TagDefinition"):
        TagDefinition("asdf://somewhere.org/extensions/foo/tags/foo-*")


def test_converter():
    class ConverterNoSubclass:
        tags = []
        types = []

        def to_yaml_tree(self, *args):
            pass

        def from_yaml_tree(self, *args):
            pass

    assert issubclass(ConverterNoSubclass, Converter)


def test_converter_proxy():
    # Test the minimum set of converter methods:
    extension = ExtensionProxy(MinimumExtension())
    converter = MinimumConverter()
    proxy = ConverterProxy(converter, extension)

    assert isinstance(proxy, Converter)

    assert proxy.tags == []
    assert proxy.types == []
    assert proxy.to_yaml_tree(None, None, None) == "to_yaml_tree result"
    assert proxy.from_yaml_tree(None, None, None) == "from_yaml_tree result"
    assert proxy.tags == []
    assert proxy.delegate is converter
    assert proxy.extension == extension
    assert proxy.package_name is None
    assert proxy.package_version is None
    assert proxy.class_name == "asdf._tests.test_extension.MinimumConverter"

    # Check the __eq__ and __hash__ behavior:
    assert proxy == ConverterProxy(converter, extension)
    assert proxy != ConverterProxy(MinimumConverter(), extension)
    assert proxy != ConverterProxy(converter, MinimumExtension())
    assert proxy in {ConverterProxy(converter, extension)}
    assert proxy not in {ConverterProxy(MinimumConverter(), extension), ConverterProxy(converter, MinimumExtension())}

    # Check the __repr__:
    assert "class: asdf._tests.test_extension.MinimumConverter" in repr(proxy)
    assert "package: (none)" in repr(proxy)

    # Test the full set of converter methods:
    converter = FullConverter(
        tags=[
            "asdf://somewhere.org/extensions/test/tags/foo-*",
            "asdf://somewhere.org/extensions/test/tags/bar-*",
        ],
        types=[FooType, BarType],
    )

    extension = FullExtension(
        tags=[
            TagDefinition(
                "asdf://somewhere.org/extensions/test/tags/foo-1.0",
                schema_uris="asdf://somewhere.org/extensions/test/schemas/foo-1.0",
                title="Foo tag title",
                description="Foo tag description",
            ),
            TagDefinition(
                "asdf://somewhere.org/extensions/test/tags/bar-1.0",
                schema_uris="asdf://somewhere.org/extensions/test/schemas/bar-1.0",
                title="Bar tag title",
                description="Bar tag description",
            ),
        ],
    )

    extension_proxy = ExtensionProxy(extension, package_name="foo", package_version="1.2.3")
    proxy = ConverterProxy(converter, extension_proxy)
    assert len(proxy.tags) == 2
    assert "asdf://somewhere.org/extensions/test/tags/foo-1.0" in proxy.tags
    assert "asdf://somewhere.org/extensions/test/tags/bar-1.0" in proxy.tags
    assert proxy.types == [FooType, BarType]
    assert proxy.to_yaml_tree(None, None, None) == "to_yaml_tree result"
    assert proxy.from_yaml_tree(None, None, None) == "from_yaml_tree result"
    assert proxy.select_tag(None, None) == "select_tag result"
    assert proxy.delegate is converter
    assert proxy.extension == extension_proxy
    assert proxy.package_name == "foo"
    assert proxy.package_version == "1.2.3"
    assert proxy.class_name == "asdf._tests.test_extension.FullConverter"

    # Check the __repr__ since it will contain package info now:
    assert "class: asdf._tests.test_extension.FullConverter" in repr(proxy)
    assert "package: foo==1.2.3" in repr(proxy)

    # Should error because object() does fulfill the Converter interface:
    with pytest.raises(TypeError, match=r"Converter must implement the .*"):
        ConverterProxy(object(), extension)

    # Should fail because tags must be str:
    with pytest.raises(TypeError, match=r"Converter property .* must contain str values"):
        ConverterProxy(MinimumConverter(tags=[object()]), extension)

    # Should fail because types must instances of type:
    with pytest.raises(TypeError, match=r"Converter property .* must contain str or type values"):
        # as the code will ignore types if no relevant tags are found
        # include a tag from this extension to make sure the proxy considers
        # the types
        ConverterProxy(MinimumConverter(tags=[extension.tags[0].tag_uri], types=[object()]), extension)


def test_converter_subclass_with_no_supported_tags():
    """
    Adding a Converter to an Extension that doesn't list support for the tags
    associated with the Converter should result in a failure to convert.
    """

    class Foo:
        pass

    class FooConverterWithSubclass(Converter):
        tags = ["asdf://somewhere.org/tags/foo-1.0.0"]
        types = [Foo]

        def to_yaml_tree(self, *args):
            pass

        def from_yaml_tree(self, *args):
            pass

    class FooExtension(Extension):
        tags = []
        converters = [FooConverterWithSubclass()]
        extension_uri = "asdf://somewhere.org/extensions/foo-1.0.0"

    tree = {"obj": Foo()}
    with config_context() as cfg:
        cfg.add_extension(FooExtension())
        with pytest.raises(AsdfSerializationError, match=r"is not serializable by asdf"):
            roundtrip_object(tree)


def test_manifest_extension():
    with config_context() as config:
        minimal_manifest = """%YAML 1.1
---
id: asdf://somewhere.org/manifests/foo
extension_uri: asdf://somewhere.org/extensions/foo
...
"""
        config.add_resource_mapping({"asdf://somewhere.org/extensions/foo": minimal_manifest})
        extension = ManifestExtension.from_uri("asdf://somewhere.org/extensions/foo")
        assert isinstance(extension, Extension)
        assert extension.extension_uri == "asdf://somewhere.org/extensions/foo"
        assert extension.legacy_class_names == []
        assert extension.asdf_standard_requirement is None
        assert extension.converters == []
        assert extension.compressors == []
        assert extension.validators == []
        assert extension.tags == []

        proxy = ExtensionProxy(extension)
        assert proxy.extension_uri == "asdf://somewhere.org/extensions/foo"
        assert proxy.legacy_class_names == set()
        assert proxy.asdf_standard_requirement == SpecifierSet()
        assert proxy.converters == []
        assert proxy.compressors == []
        assert proxy.validators == []
        assert proxy.tags == []

    with config_context() as config:
        full_manifest = """%YAML 1.1
---
id: asdf://somewhere.org/manifests/foo
extension_uri: asdf://somewhere.org/extensions/foo
asdf_standard_requirement:
  gte: 1.6.0
  lt: 2.0.0
tags:
  - asdf://somewhere.org/tags/bar
  - tag_uri: asdf://somewhere.org/tags/baz
    schema_uri: asdf://somewhere.org/schemas/baz
    title: Baz title
    description: Bar description
...
"""
        config.add_resource_mapping({"asdf://somewhere.org/extensions/foo": full_manifest})

        class FooConverter:
            tags = ["asdf://somewhere.org/tags/bar", "asdf://somewhere.org/tags/baz"]
            types = []

            def select_tag(self, *args):
                pass

            def to_yaml_tree(self, *args):
                pass

            def from_yaml_tree(self, *args):
                pass

        converter = FooConverter()
        validator = MinimalValidator()
        compressor = MinimalCompressor()

        extension = ManifestExtension.from_uri(
            "asdf://somewhere.org/extensions/foo",
            legacy_class_names=["foo.extension.LegacyExtension"],
            converters=[converter],
            compressors=[compressor],
            validators=[validator],
        )
        assert extension.extension_uri == "asdf://somewhere.org/extensions/foo"
        assert extension.legacy_class_names == ["foo.extension.LegacyExtension"]
        assert extension.asdf_standard_requirement == SpecifierSet(">=1.6.0,<2.0.0")
        assert extension.converters == [converter]
        assert extension.compressors == [compressor]
        assert extension.validators == [validator]
        assert len(extension.tags) == 2
        assert extension.tags[0] == "asdf://somewhere.org/tags/bar"
        assert extension.tags[1].tag_uri == "asdf://somewhere.org/tags/baz"
        assert extension.tags[1].schema_uris == ["asdf://somewhere.org/schemas/baz"]
        assert extension.tags[1].title == "Baz title"
        assert extension.tags[1].description == "Bar description"

        proxy = ExtensionProxy(extension)
        assert proxy.extension_uri == "asdf://somewhere.org/extensions/foo"
        assert proxy.legacy_class_names == {"foo.extension.LegacyExtension"}
        assert proxy.asdf_standard_requirement == SpecifierSet(">=1.6.0,<2.0.0")
        assert proxy.converters == [ConverterProxy(converter, proxy)]
        assert proxy.compressors == [compressor]
        assert proxy.validators == [validator]
        assert len(proxy.tags) == 2
        assert proxy.tags[0].tag_uri == "asdf://somewhere.org/tags/bar"
        assert proxy.tags[1].tag_uri == "asdf://somewhere.org/tags/baz"
        assert proxy.tags[1].schema_uris == ["asdf://somewhere.org/schemas/baz"]
        assert proxy.tags[1].title == "Baz title"
        assert proxy.tags[1].description == "Bar description"

    with config_context() as config:
        simple_asdf_standard_manifest = """%YAML 1.1
---
id: asdf://somewhere.org/manifests/foo
extension_uri: asdf://somewhere.org/extensions/foo
asdf_standard_requirement: 1.6.0
...
"""
        config.add_resource_mapping({"asdf://somewhere.org/extensions/foo": simple_asdf_standard_manifest})
        extension = ManifestExtension.from_uri("asdf://somewhere.org/extensions/foo")
        assert extension.asdf_standard_requirement == SpecifierSet("==1.6.0")

        proxy = ExtensionProxy(extension)
        assert proxy.asdf_standard_requirement == SpecifierSet("==1.6.0")


def test_validator():
    validator = MinimalValidator()
    extension = FullExtension(validators=[validator])

    failing_schema = """
        type: object
        properties:
          foo:
            fail: true
    """

    passing_schema = """
        type: object
        properties:
          foo:
            fail: false
    """

    with config_context() as config:
        config.add_extension(extension)
        config.add_resource_mapping(
            {
                "asdf://somewhere.org/schemas/failing": failing_schema,
                "asdf://somewhere.org/schemas/passing": passing_schema,
            },
        )

        with AsdfFile(custom_schema="asdf://somewhere.org/schemas/passing") as af:
            af["foo"] = "bar"
            af.validate()

        with AsdfFile(custom_schema="asdf://somewhere.org/schemas/failing") as af:
            af.validate()

            af["foo"] = "bar"
            with pytest.raises(ValidationError, match=r"Node was doomed to fail"):
                af.validate()


def test_converter_deferral():
    class Bar:
        def __init__(self, value):
            self.value = value

    class Foo(Bar):
        pass

    class Baz(Bar):
        pass

    class FooConverter:
        tags = []
        types = [Foo]

        def select_tag(self, *args):
            return None

        def to_yaml_tree(self, obj, tag, ctx):
            # convert Foo instance to Bar
            return Bar(obj.value)

        def from_yaml_tree(self, node, tag, ctx):
            raise NotImplementedError()

    class BarConverter:
        tags = ["asdf://somewhere.org/tags/bar"]
        types = [Bar]

        def to_yaml_tree(self, obj, tag, ctx):
            return {"value": obj.value}

        def from_yaml_tree(self, node, tag, ctx):
            return Bar(node["value"])

    class BazConverter:
        tags = []
        types = [Baz]

        def select_tag(self, *args):
            return None

        def to_yaml_tree(self, obj, tag, ctx):
            return Foo(obj.value)

        def from_yaml_tree(self, node, tag, ctx):
            raise NotImplementedError()

    extension = FullExtension(converters=[FooConverter(), BarConverter(), BazConverter()], tags=BarConverter.tags)
    with config_context() as config:
        config.add_extension(extension)

        foo = Foo(26)
        bar = Bar(42)
        baz = Baz(720)

        bar_rt = roundtrip_object(bar)
        assert isinstance(bar_rt, Bar)
        assert bar_rt.value == bar.value

        foo_rt = roundtrip_object(foo)
        assert isinstance(foo_rt, Bar)
        assert foo_rt.value == foo.value

        baz_rt = roundtrip_object(baz)
        assert isinstance(baz_rt, Bar)
        assert baz_rt.value == baz.value


def test_converter_loop():
    class Bar:
        def __init__(self, value):
            self.value = value

    class Foo(Bar):
        pass

    class Baz(Bar):
        pass

    class FooConverter:
        tags = []
        types = [Foo]

        def select_tag(self, *args):
            return None

        def to_yaml_tree(self, obj, tag, ctx):
            return Bar(obj.value)

        def from_yaml_tree(self, node, tag, ctx):
            raise NotImplementedError()

    class BarConverter:
        tags = []
        types = [Bar]

        def select_tag(self, *args):
            return None

        def to_yaml_tree(self, obj, tag, ctx):
            return Baz(obj.value)

        def from_yaml_tree(self, node, tag, ctx):
            raise NotImplementedError()

    class BazConverter:
        tags = []
        types = [Baz]

        def select_tag(self, *args):
            return None

        def to_yaml_tree(self, obj, tag, ctx):
            return Foo(obj.value)

        def from_yaml_tree(self, node, tag, ctx):
            raise NotImplementedError()

    extension = FullExtension(converters=[FooConverter(), BarConverter(), BazConverter()])
    with config_context() as config:
        config.add_extension(extension)

        for typ in (Foo, Bar, Baz):
            obj = typ(42)
            with pytest.raises(TypeError, match=r"Conversion cycle detected"):
                roundtrip_object(obj)


@pytest.mark.parametrize("is_subclass", [True, False])
@pytest.mark.parametrize("indirect", [True, False])
def test_warning_or_error_for_default_select_tag(is_subclass, indirect):
    class Foo:
        pass

    ParentClass = Converter if is_subclass else object

    if indirect:

        class IntermediateClass(ParentClass):
            pass

        ParentClass = IntermediateClass

    class FooConverter(ParentClass):
        tags = ["asdf://somewhere.org/tags/foo-*"]
        types = [Foo]

        def to_yaml_tree(self, obj, tag, ctx):
            return {}

        def from_yaml_tree(self, node, tag, ctx):
            return Foo()

    tags = [
        "asdf://somewhere.org/tags/foo-1.0.0",
        "asdf://somewhere.org/tags/foo-2.0.0",
    ]
    extension = FullExtension(converters=[FooConverter()], tags=tags)
    with config_context() as config:
        with pytest.raises(RuntimeError, match="Converter handles multiple tags"):
            config.add_extension(extension)


def test_reference_cycle(tmp_path, with_lazy_tree):
    class FractionWithInverse(fractions.Fraction):
        def __init__(self, *args, **kwargs):
            self._inverse = None

        @property
        def inverse(self):
            return self._inverse

        @inverse.setter
        def inverse(self, value):
            self._inverse = value

    class FractionWithInverseConverter:
        tags = ["asdf://example.com/fractions/tags/fraction-1.0.0"]
        types = [FractionWithInverse]

        def to_yaml_tree(self, obj, tag, ctx):
            return {
                "numerator": obj.numerator,
                "denominator": obj.denominator,
                "inverse": obj.inverse,
            }

        def from_yaml_tree(self, node, tag, ctx):
            obj = FractionWithInverse(node["numerator"], node["denominator"])
            yield obj
            obj.inverse = node["inverse"]

    class FractionWithInverseExtension:
        tags = FractionWithInverseConverter.tags
        converters = [FractionWithInverseConverter()]
        extension_uri = "asdf://example.com/fractions/extensions/fraction-1.0.0"

    with config_context() as cfg:
        cfg.add_extension(FractionWithInverseExtension())

        f1 = FractionWithInverse(3, 5)
        f2 = FractionWithInverse(5, 3)
        f1.inverse = f2
        f2.inverse = f1
        fn = tmp_path / "test.asdf"
        asdf.AsdfFile({"obj": f1}).write_to(fn)
        with asdf.open(fn) as af:
            read_f1 = af["obj"]
            assert read_f1.inverse.inverse is read_f1


def test_manifest_uri_id_mismatch_warning(tmp_path):
    with config_context() as config:
        # make an extension with a manifest (where id doesn't match the registered uri)
        full_manifest = """%YAML 1.1
---
id: asdf://somewhere.org/manifests/foo
extension_uri: asdf://somewhere.org/extensions/foo
tags:
  - asdf://somewhere.org/tags/bar
...
"""
        config.add_resource_mapping({"asdf://somewhere.org/extensions/foo": full_manifest})

        class Foo:
            pass

        class FooConverter:
            tags = ["asdf://somewhere.org/tags/bar"]
            types = [Foo]

            def to_yaml_tree(self, *args):
                return {}

            def from_yaml_tree(self, *args):
                return Foo()

        extension = ManifestExtension.from_uri(
            "asdf://somewhere.org/extensions/foo",
            converters=[FooConverter()],
        )

        # use the extension to write out a file
        config.add_extension(extension)

        af = AsdfFile()
        af["foo"] = Foo()
        fn = tmp_path / "foo.asdf"
        with pytest.warns(AsdfManifestURIMismatchWarning):
            af.write_to(fn)


def test_resolve_type_not_imported():
    path = "mailbox.Mailbox"

    if "mailbox" in sys.modules:
        del sys.modules["mailbox"]

    assert _resolve_type(path) is None

    import mailbox

    assert _resolve_type(path) is mailbox.Mailbox


@pytest.mark.parametrize(
    "path, obj", (("sys", sys), ("asdf.AsdfFile", AsdfFile), ("asdf.Missing", None), ("not_a_module", None))
)
def test_resolve_type(path, obj):
    assert _resolve_type(path) is obj


def test_extension_converter_by_class_path():
    class MailboxConverter:
        tags = ["asdf://example.com/tags/mailbox-1.0.0"]
        types = ["mailbox.Mailbox"]

        def to_yaml_tree(self, obj, tag, ctx):
            return {}

        def from_yaml_tree(self, node, tag, ctx):
            return None

    class MailboxExtension:
        tags = MailboxConverter.tags
        converters = [MailboxConverter()]
        extension_uri = "asdf://example.com/extensions/mailbox-1.0.0"

    # grab the type so we can use it for extension_manager.get_converter_for_type
    import mailbox

    typ = mailbox.Mailbox
    del sys.modules["mailbox"], mailbox

    with config_context() as cfg:
        cfg.add_extension(MailboxExtension())
        extension_manager = AsdfFile().extension_manager

        # make sure that registering the extension did not load the module
        assert "mailbox" not in sys.modules

        # as the module hasn't been loaded, the converter shouldn't be found
        with pytest.raises(KeyError, match=r"No support available for Python type 'mailbox\.Mailbox'"):
            extension_manager.get_converter_for_type(typ)

        # make sure inspecting the type didn't import the module
        assert "mailbox" not in sys.modules

        # finally, import the module and check that the converter can now be found
        import mailbox

        converter = extension_manager.get_converter_for_type(mailbox.Mailbox)
        assert isinstance(converter.delegate, MailboxConverter)


def test_named_tuple_extension():
    Point = collections.namedtuple("Point", ["x", "y"])

    class PointConverter:
        tags = ["asdf://example.com/tags/point-1.0.0"]
        types = [Point]

        def to_yaml_tree(self, obj, tag, ctx):
            return list(obj)

        def from_yaml_tree(self, node, tag, ctx):
            return Point(*node)

    class PointExtension:
        tags = PointConverter.tags
        converters = [PointConverter()]
        extension_uri = "asdf://example.com/extensions/point-1.0.0"

    pt = Point(1, 2)

    # without the extension we can't serialize this
    with pytest.raises(AsdfSerializationError, match="is not serializable by asdf"):
        roundtrip_object(pt)

    with config_context() as cfg:
        cfg.add_extension(PointExtension())
        rpt = roundtrip_object(pt)
        assert isinstance(rpt, Point)



import numpy as np

import asdf


def test_writes_but_fails_to_read_inline_structured_array(tmp_path):
    """
    ASDF writes but fails to read inline structured array

    https://github.com/asdf-format/asdf/issues/1540
    """
    x = np.array((0, 1.0, [2, 3]), dtype=[("MINE", "i1"), ("f1", "<f8"), ("arr", "<i4", (2,))])
    af = asdf.AsdfFile()
    af["x"] = x
    af.write_to("test.asdf", all_array_storage="inline")

    with asdf.open("test.asdf") as af2:
        np.testing.assert_array_equal(af2["x"], x)



from . import versioning
from ._version import version as asdf_package_version


def validate_version(version):
    # Account for the possibility of AsdfVersion
    version = str(version)
    if version not in versioning.supported_versions:
        msg = "ASDF Standard version {} is not supported by asdf=={}.  Available ASDF Standard versions: {}".format(
            version,
            asdf_package_version,
            ", ".join(str(v) for v in versioning.supported_versions),
        )
        raise ValueError(msg)
    return version



import contextlib
import io
import os
import re
import sys

import numpy as np
import pytest
import yaml
from numpy import ma
from numpy.testing import assert_array_equal

import asdf
from asdf.exceptions import ValidationError
from asdf.extension import Converter, Extension, TagDefinition
from asdf.tags.core import ndarray
from asdf.testing import helpers


@pytest.fixture
def ndarray_tag():
    af = asdf.AsdfFile()
    cvt = af.extension_manager.get_converter_for_type(np.ndarray)
    full_tag = cvt.select_tag(np.zeros(0), af)
    return full_tag.removeprefix("tag:stsci.edu:asdf/")


# These custom types and the custom extension are here purely for the purpose
# of testing NDArray objects and making sure that they can be validated as part
# of a nested hierarchy, and not just top-level objects.
class CustomData:
    def __init__(self, value):
        self.value = value


class CustomNDim:
    def __init__(self, value):
        self.value = value


class CustomNDimConverter(Converter):
    tags = ["tag:nowhere.org:custom/ndim-1.0.0"]
    types = [CustomNDim]

    def to_yaml_tree(self, obj, tag, ctx):
        return obj.value

    def from_yaml_tree(self, node, tag, ctx):
        return CustomNDim(node)


class CustomDataConverter(Converter):
    tags = ["tag:nowhere.org:custom/datatype-1.0.0"]
    types = [CustomData]

    def to_yaml_tree(self, obj, tag, ctx):
        return obj.value

    def from_yaml_tree(self, node, tag, ctx):
        return CustomData(node)


class CustomExtension(Extension):
    tags = [
        TagDefinition(
            tag_uri="tag:nowhere.org:custom/datatype-1.0.0",
            schema_uris=["http://nowhere.org/schemas/custom/datatype-1.0.0"],
        ),
        TagDefinition(
            tag_uri="tag:nowhere.org:custom/ndim-1.0.0",
            schema_uris=["http://nowhere.org/schemas/custom/ndim-1.0.0"],
        ),
    ]
    extension_uri = "asdf://nowhere.org/extensions/custom-1.0.0"
    converters = [CustomDataConverter(), CustomNDimConverter()]


@contextlib.contextmanager
def with_custom_extension():
    with asdf.config_context() as cfg:
        cfg.add_extension(CustomExtension())
        cfg.add_resource_mapping(
            {
                "http://nowhere.org/schemas/custom/datatype-1.0.0": """%YAML 1.1
---
$schema: "http://stsci.edu/schemas/asdf/asdf-schema-1.1.0"
id: "http://nowhere.org/schemas/custom/datatype-1.0.0"
type: object
properties:
  a:
    datatype: float32

  b:
    datatype: float32
    exact_datatype: true

  c:
    datatype:
      - name: a
        datatype: int16
      - name: b
        datatype: ['ascii', 16]

  d:
    datatype:
      - name: a
        datatype: int16
      - name: b
        datatype: ['ascii', 16]
    exact_datatype: true
  e:
    anyOf:
      - type: "null"
      - datatype: int16""",
                "http://nowhere.org/schemas/custom/ndim-1.0.0": """%YAML 1.1
---
$schema: "http://stsci.edu/schemas/asdf/asdf-schema-1.1.0"
id: "http://nowhere.org/schemas/custom/ndim-1.0.0"
type: object
properties:
  a:
    ndim: 2

  b:
    max_ndim: 2""",
            }
        )
        yield


@contextlib.contextmanager
def roundtrip(af, raw=False, standard_version=None):
    if not isinstance(af, asdf.AsdfFile):
        af = asdf.AsdfFile(af, version=standard_version)
    b = io.BytesIO()
    af.write_to(b)
    b.seek(0)
    if raw:
        bs = b.read()
        if asdf.constants.BLOCK_MAGIC in bs:
            bs, *_ = bs.split(asdf.constants.BLOCK_MAGIC)
        yield bs
    else:
        with asdf.open(b) as af:
            yield af


def test_sharing():
    x = np.arange(0, 10, dtype=float)
    tree = {"science_data": x, "subset": x[3:-3], "skipping": x[::2]}

    with roundtrip(tree) as af:
        tree = af.tree
        assert_array_equal(tree["science_data"], x)
        assert_array_equal(tree["subset"], x[3:-3])
        assert_array_equal(tree["skipping"], x[::2])

        assert tree["science_data"].ctypes.data == tree["skipping"].ctypes.data

        assert len(af._blocks.blocks) == 1
        assert af._blocks.blocks[0].header["data_size"] == 80

        tree["science_data"][0] = 42
        assert tree["skipping"][0] == 42


def test_byteorder():
    tree = {
        "bigendian": np.arange(0, 10, dtype=">f8"),
        "little": np.arange(0, 10, dtype="<f8"),
    }

    with roundtrip(tree) as af:
        my_tree = af.tree
        for endian in ("bigendian", "little"):
            assert my_tree[endian].dtype == tree[endian].dtype

        if sys.byteorder == "little":
            assert my_tree["bigendian"].dtype.byteorder == ">"
            assert my_tree["little"].dtype.byteorder == "="
        else:
            assert my_tree["bigendian"].dtype.byteorder == "="
            assert my_tree["little"].dtype.byteorder == "<"


@pytest.mark.parametrize("dtype", ndarray._datatype_names.values())
def test_all_dtypes(dtype):
    standard_version = "1.6.0" if dtype == "f2" else None
    tree = {}
    for byteorder in (">", "<"):
        arr = np.array([True, False]) if dtype == "b1" else np.arange(0, 10, dtype=str(byteorder + dtype))
        tree[byteorder + dtype] = arr
    with roundtrip(tree, standard_version=standard_version) as af:
        for k in tree:
            pre = tree[k]
            post = af[k]
            assert_array_equal(pre, post)


def test_dont_load_data():
    x = np.arange(0, 10, dtype=float)
    tree = {"science_data": x, "subset": x[3:-3], "skipping": x[::2]}
    ff = asdf.AsdfFile(tree)

    buff = io.BytesIO()
    ff.write_to(buff)

    buff.seek(0)
    with asdf.open(buff) as ff:
        # repr and str shouldn't load data
        str(ff.tree["science_data"])
        repr(ff.tree)

        for block in ff._blocks.blocks:
            assert callable(block._data)


def test_table_inline():
    table = np.array(
        [(0, 1, (2, 3)), (4, 5, (6, 7))],
        dtype=[("MINE", np.int8), ("", np.float64), ("arr", ">i4", (2,))],
    )

    tree = {"table_data": table}

    with asdf.config_context() as config:
        config.array_inline_threshold = 100

        with roundtrip(tree) as af:
            assert table.dtype.names == af["table_data"].dtype.names
            for n in table.dtype.names:
                assert_array_equal(table[n], af["table_data"][n])

        with roundtrip(tree, raw=True) as content:
            tree = yaml.safe_load(re.sub(rb"!core/\S+", b"", content))

            assert tree["table_data"] == {
                "datatype": [
                    {"datatype": "int8", "name": "MINE"},
                    {"datatype": "float64", "name": "f1"},
                    {"datatype": "int32", "name": "arr", "shape": [2]},
                ],
                "data": [[0, 1.0, [2, 3]], [4, 5.0, [6, 7]]],
                "shape": [2],
            }


def test_array_inline_threshold_recursive(tmp_path):
    """
    Test that setting the inline threshold works for objects
    that contain (and when serialized produce a ndarray)
    """

    class NDArrayContainer:
        def __init__(self, array):
            self._array = array

        @property
        def array(self):
            return np.array(self._array)

    class NDArrayContainerConverter:
        tags = ["http://somewhere.org/tags/foo-1.0.0"]
        types = [NDArrayContainer]

        def to_yaml_tree(self, obj, tag, ctx):
            return {"array": obj.array}

        def from_yaml_tree(self, node, tag, ctx):
            return NDArrayContainer(node["array"])

    class NDArrayContainerExtension:
        tags = NDArrayContainerConverter.tags
        converters = [NDArrayContainerConverter()]
        extension_uri = "http://somewhere.org/extensions/foo-1.0.0"

    container = NDArrayContainer([[1, 2], [3, 4]])
    tree = {"test": container}

    with asdf.config_context() as config:
        config.add_extension(NDArrayContainerExtension())
        config.array_inline_threshold = 100
        # we can no longer use _helpers.assert_roundtrip_tree here because
        # the model no longer has a CustomType which results in equality testing
        # using == which will fail
        # this test appears to be designed to test the inline threshold so we can
        # just look at the number of blocks
        fn = str(tmp_path / "test.asdf")
        af = asdf.AsdfFile(tree)
        af.write_to(fn)
        with asdf.open(fn) as af:
            assert len(list(af._blocks.blocks)) == 0


def test_copy_inline(ndarray_tag):
    yaml = f"""
x0: !{ndarray_tag}
  data: [-1.0, 1.0]
    """

    buff = helpers.yaml_to_asdf(yaml)

    with asdf.open(buff) as infile, asdf.AsdfFile() as f:
        f.tree["a"] = infile.tree["x0"]
        f.tree["b"] = f.tree["a"]
        f.write_to(io.BytesIO())


def test_table():
    table = np.array([(0, 1, (2, 3)), (4, 5, (6, 7))], dtype=[("MINE", np.int8), ("", "<f8"), ("arr", ">i4", (2,))])

    tree = {"table_data": table}

    with roundtrip(tree) as af:
        assert_array_equal(af["table_data"], table)

    with roundtrip(tree, raw=True) as content:
        tree = yaml.safe_load(re.sub(rb"!core/\S+", b"", content))

        assert tree["table_data"] == {
            "datatype": [
                {"byteorder": "big", "datatype": "int8", "name": "MINE"},
                {"byteorder": "little", "datatype": "float64", "name": "f1"},
                {"byteorder": "big", "datatype": "int32", "name": "arr", "shape": [2]},
            ],
            "shape": [2],
            "source": 0,
            "byteorder": "big",
        }


def test_table_nested_fields():
    table = np.array(
        [(0, (1, 2)), (4, (5, 6)), (7, (8, 9))],
        dtype=[("A", "<i8"), ("B", [("C", "<i8"), ("D", "<i8")])],
    )

    tree = {"table_data": table}

    with roundtrip(tree) as af:
        assert_array_equal(af["table_data"], table)

    with roundtrip(tree, raw=True) as content:
        tree = yaml.safe_load(re.sub(rb"!core/\S+", b"", content))

        assert tree["table_data"] == {
            "datatype": [
                {"datatype": "int64", "name": "A", "byteorder": "little"},
                {
                    "datatype": [
                        {"datatype": "int64", "name": "C", "byteorder": "little"},
                        {"datatype": "int64", "name": "D", "byteorder": "little"},
                    ],
                    "name": "B",
                    "byteorder": "big",
                },
            ],
            "shape": [3],
            "source": 0,
            "byteorder": "big",
        }


def test_inline():
    x = np.arange(0, 10, dtype=float)
    tree = {"science_data": x, "subset": x[3:-3], "skipping": x[::2]}

    buff = io.BytesIO()

    ff = asdf.AsdfFile(tree)
    ff.set_array_storage(x, "inline")
    ff.write_to(buff)

    buff.seek(0)
    with asdf.open(buff, mode="rw") as ff:
        for k in tree:
            assert_array_equal(tree[k], ff.tree[k])
        assert len(list(ff._blocks.blocks)) == 0
        buff = io.BytesIO()
        ff.write_to(buff)

    assert b"[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]" in buff.getvalue()


def test_inline_bare(ndarray_tag):
    content = f"arr: !{ndarray_tag} [[1, 2, 3, 4], [5, 6, 7, 8]]"
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff) as ff:
        assert_array_equal(ff.tree["arr"], [[1, 2, 3, 4], [5, 6, 7, 8]])


@pytest.mark.parametrize(
    "mask",
    [
        [[False, False, True], [False, True, False], [False, False, False]],
        True,
        False,
    ],
)
def test_mask_roundtrip(mask):
    array = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 0]])
    tree = {
        "unmasked": array,
        "masked": np.ma.array(array, mask=mask),
    }

    with roundtrip(tree) as af:
        # assert_array_equal ignores the mask, so use equality here
        assert (tree["masked"].data == af["masked"].data).all()
        assert (tree["masked"].mask == af["masked"].mask).all()
        # ensure tree validity
        assert (af["unmasked"] == af["masked"].data).all()
        assert len(af._blocks.blocks) == 2


def test_len_roundtrip():
    sequence = np.arange(0, 10, dtype=int)
    tree = {"sequence": sequence}

    with roundtrip(tree) as af:
        s = af.tree["sequence"]
        assert len(s) == 10


def test_mask_arbitrary(ndarray_tag):
    content = f"""
arr: !{ndarray_tag}
  data: [[1, 2, 3, 1234], [5, 6, 7, 8]]
  mask: 1234
    """

    buff = helpers.yaml_to_asdf(content)
    with asdf.open(buff) as ff:
        assert_array_equal(ff.tree["arr"].mask, [[False, False, False, True], [False, False, False, False]])


def test_mask_nan(ndarray_tag):
    content = f"""
arr: !{ndarray_tag}
  data: [[1, 2, 3, .NaN], [5, 6, 7, 8]]
  mask: .NaN
    """

    buff = helpers.yaml_to_asdf(content)
    with asdf.open(buff) as ff:
        assert_array_equal(ff.tree["arr"].mask, [[False, False, False, True], [False, False, False, False]])


def test_string():
    tree = {
        "ascii": np.array([b"foo", b"bar", b"baz"]),
        "unicode": np.array(["სამეცნიერო", "данные", "வடிவம்"]),
    }

    with roundtrip(tree) as af:
        for k in tree:
            assert_array_equal(tree[k], af[k])


def test_string_table():
    tree = {"table": np.array([(b"foo", "სამეცნიერო", "42", "53.0")])}

    with roundtrip(tree) as af:
        for k in tree:
            assert_array_equal(tree[k], af[k])


def test_inline_string(ndarray_tag):
    content = f"arr: !{ndarray_tag} ['a', 'b', 'c']"
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff) as ff:
        assert_array_equal(ff.tree["arr"]._make_array(), ["a", "b", "c"])


def test_inline_structured(ndarray_tag):
    content = f"""
arr: !{ndarray_tag}
    datatype: [['ascii', 4], uint16, uint16, ['ascii', 4]]
    data: [[M110, 110, 205, And],
           [ M31,  31, 224, And],
           [ M32,  32, 221, And],
           [M103, 103, 581, Cas]]"""

    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff) as ff:
        assert ff.tree["arr"]["f1"].dtype.char == "H"


def test_simple_table():
    table = np.array(
        [
            (10.683262825012207, 41.2674560546875, 0.13, 0.12, 213.916),
            (10.682777404785156, 41.270111083984375, 0.1, 0.09, 306.825),
            (10.684737205505371, 41.26903533935547, 0.08, 0.07, 96.656),
            (10.682382583618164, 41.26792526245117, 0.1, 0.09, 237.145),
            (10.686025619506836, 41.26922607421875, 0.13, 0.12, 79.581),
            (10.685656547546387, 41.26955032348633, 0.13, 0.12, 55.219),
            (10.684028625488281, 41.27090072631836, 0.13, 0.12, 345.269),
            (10.687610626220703, 41.270301818847656, 0.18, 0.14, 60.192),
        ],
        dtype=[
            ("ra", "<f4"),
            ("dec", "<f4"),
            ("err_maj", "<f8"),
            ("err_min", "<f8"),
            ("angle", "<f8"),
        ],
    )

    tree = {"table": table}
    ff = asdf.AsdfFile(tree)
    ff.set_array_storage(table, "inline")
    ff.write_to(io.BytesIO())


def test_unicode_to_list():
    arr = np.array(["", "𐀠"], dtype="<U")
    tree = {"unicode": arr}

    fd = io.BytesIO()
    ff = asdf.AsdfFile(tree)
    ff.set_array_storage(arr, "inline")
    ff.write_to(fd)
    fd.seek(0)

    with asdf.open(fd) as ff:
        ff.resolve_references()
        ff.write_to(io.BytesIO(), all_array_storage="inline")


def test_inline_masked_array(tmp_path):
    testfile = os.path.join(str(tmp_path), "masked.asdf")

    tree = {"test": ma.array([1, 2, 3], mask=[0, 1, 0])}

    f = asdf.AsdfFile(tree)
    f.set_array_storage(tree["test"], "inline")
    f.write_to(testfile)

    with asdf.open(testfile) as f2:
        assert len(list(f2._blocks.blocks)) == 0
        # assert_array_equal ignores the mask, so use equality here
        assert (f.tree["test"] == f2.tree["test"]).all()

    with open(testfile, "rb") as fd:
        assert b"null" in fd.read()


def test_masked_array_stay_open_bug(tmp_path):
    psutil = pytest.importorskip("psutil")

    tmppath = os.path.join(str(tmp_path), "masked.asdf")

    tree = {"test": np.ma.array([1, 2, 3], mask=[False, True, False])}

    f = asdf.AsdfFile(tree)
    f.write_to(tmppath)

    p = psutil.Process()
    orig_open = p.open_files()

    for _ in range(3):
        with asdf.open(tmppath) as f2:
            np.sum(f2.tree["test"])

    assert len(p.open_files()) <= len(orig_open)


def test_memmap_stay_open_bug(tmp_path):
    """
    Regression test for issue #1239
    memmapped arrays only closed at garbage collection when asdf.open given an open file

    When asdf.open is called with an already opened file
    it did not close any memmaps that it created (as the file
    pointer was still valid). These lingered until garbage collection
    and caused CI failures in astropy:
    https://github.com/astropy/astropy/pull/14035#issuecomment-1325236928
    """
    psutil = pytest.importorskip("psutil")

    tmppath = os.path.join(str(tmp_path), "arr.asdf")

    tree = {"test": np.array([1, 2, 3])}

    f = asdf.AsdfFile(tree)
    f.write_to(tmppath)

    p = psutil.Process()
    orig_open = p.open_files()

    for _ in range(3):
        with open(tmppath, mode="rb") as fp, asdf.open(fp) as f2:
            np.sum(f2.tree["test"])

    assert len(p.open_files()) <= len(orig_open)


def test_masked_array_repr(tmp_path):
    tmppath = os.path.join(str(tmp_path), "masked.asdf")

    tree = {"array": np.arange(10), "masked": np.ma.array([1, 2, 3], mask=[False, True, False])}

    asdf.AsdfFile(tree).write_to(tmppath)

    with asdf.open(tmppath) as ff:
        assert "masked array" in repr(ff.tree["masked"])


def test_operations_on_ndarray_proxies(tmp_path):
    tmppath = os.path.join(str(tmp_path), "test.asdf")

    tree = {"array": np.arange(10)}

    asdf.AsdfFile(tree).write_to(tmppath)

    with asdf.open(tmppath) as ff:
        x = ff.tree["array"] * 2
        assert_array_equal(x, np.arange(10) * 2)

    with asdf.open(tmppath) as ff:
        x = -ff.tree["array"]
        assert_array_equal(x, -np.arange(10))

    with asdf.open(tmppath, mode="rw") as ff:
        ff.tree["array"][2] = 4
        x = np.arange(10)
        x[2] = 4
        assert_array_equal(ff.tree["array"], x)


def test_mask_datatype(ndarray_tag):
    content = f"""
arr: !{ndarray_tag}
    data: [1, 2, 3]
    dtype: int32
    mask: !{ndarray_tag}
        data: [true, true, false]
    """
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff):
        pass


def test_invalid_mask_datatype(ndarray_tag):
    content = f"""
arr: !{ndarray_tag}
    data: [1, 2, 3]
    dtype: int32
    mask: !{ndarray_tag}
        data: ['a', 'b', 'c']
    """
    buff = helpers.yaml_to_asdf(content)

    with (
        pytest.raises(ValidationError, match=r".* is not valid under any of the given schemas"),
        asdf.open(
            buff,
        ),
    ):
        pass


@with_custom_extension()
def test_ndim_validation(ndarray_tag):
    content = f"""
obj: !<tag:nowhere.org:custom/ndim-1.0.0>
    a: !{ndarray_tag}
       data: [1, 2, 3]
    """
    buff = helpers.yaml_to_asdf(content)

    with (
        pytest.raises(ValidationError, match=r"Wrong number of dimensions:.*"),
        asdf.open(
            buff,
        ),
    ):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/ndim-1.0.0>
    a: !{ndarray_tag}
       data: [[1, 2, 3]]
    """
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/ndim-1.0.0>
    a: !{ndarray_tag}
       shape: [1, 3]
       data: [[1, 2, 3]]
    """
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/ndim-1.0.0>
    b: !{ndarray_tag}
       data: [1, 2, 3]
    """
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/ndim-1.0.0>
    b: !{ndarray_tag}
       data: [[1, 2, 3]]
    """
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/ndim-1.0.0>
    b: !{ndarray_tag}
       data: [[[1, 2, 3]]]
    """
    buff = helpers.yaml_to_asdf(content)

    with (
        pytest.raises(ValidationError, match=r"Wrong number of dimensions:.*"),
        asdf.open(
            buff,
        ),
    ):
        pass


@with_custom_extension()
def test_datatype_validation(ndarray_tag):
    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    a: !{ndarray_tag}
       data: [1, 2, 3]
       datatype: float32
    """
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    a: !{ndarray_tag}
       data: [1, 2, 3]
       datatype: float64
    """
    buff = helpers.yaml_to_asdf(content)

    with (
        pytest.raises(ValidationError, match=r"Can not safely cast from .* to .*"),
        asdf.open(
            buff,
        ),
    ):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    a: !{ndarray_tag}
       data: [1, 2, 3]
       datatype: int16
    """
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    b: !{ndarray_tag}
       data: [1, 2, 3]
       datatype: int16
    """
    buff = helpers.yaml_to_asdf(content)

    with (
        pytest.raises(ValidationError, match=r"Expected datatype .*, got .*"),
        asdf.open(
            buff,
        ),
    ):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    a: !{ndarray_tag}
       data: [[1, 'a'], [2, 'b'], [3, 'c']]
       datatype:
         - name: a
           datatype: int8
         - name: b
           datatype: ['ascii', 8]
    """
    buff = helpers.yaml_to_asdf(content)

    with (
        pytest.raises(ValidationError, match=r"Expected scalar datatype .*, got .*"),
        asdf.open(
            buff,
        ),
    ):
        pass

    content = """
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    e: null
    """
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff):
        pass


@with_custom_extension()
def test_structured_datatype_validation(ndarray_tag):
    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    c: !{ndarray_tag}
       data: [[1, 'a'], [2, 'b'], [3, 'c']]
       datatype:
         - name: a
           description: a description
           datatype: int8
         - name: b
           datatype: ['ascii', 8]
    """
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    c: !{ndarray_tag}
       data: [[1, 'a'], [2, 'b'], [3, 'c']]
       datatype:
         - name: a
           datatype: int64
         - name: b
           title: a title
           datatype: ['ascii', 8]
    """
    buff = helpers.yaml_to_asdf(content)

    with (
        pytest.raises(ValidationError, match=r"Can not safely cast to expected datatype.*"),
        asdf.open(
            buff,
        ),
    ):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    c: !{ndarray_tag}
       data: [[1, 'a', 0], [2, 'b', 1], [3, 'c', 2]]
       datatype:
         - name: a
           datatype: int8
         - name: b
           datatype: ['ascii', 8]
         - name: c
           datatype: float64
    """
    buff = helpers.yaml_to_asdf(content)

    with (
        pytest.raises(ValidationError, match=r"Mismatch in number of columns:.*"),
        asdf.open(
            buff,
        ),
    ):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    c: !{ndarray_tag}
       data: [1, 2, 3]
    """
    buff = helpers.yaml_to_asdf(content)

    with (
        pytest.raises(ValidationError, match=r"Expected structured datatype.*"),
        asdf.open(
            buff,
        ),
    ):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    d: !{ndarray_tag}
       data: [[1, 'a'], [2, 'b'], [3, 'c']]
       datatype:
         - name: a
           datatype: int8
         - name: b
           datatype: ['ascii', 8]
    """
    buff = helpers.yaml_to_asdf(content)

    with (
        pytest.raises(ValidationError, match=r"Expected datatype .*, got .*"),
        asdf.open(
            buff,
        ),
    ):
        pass

    content = f"""
obj: !<tag:nowhere.org:custom/datatype-1.0.0>
    d: !{ndarray_tag}
       data: [[1, 'a'], [2, 'b'], [3, 'c']]
       datatype:
         - name: a
           datatype: int16
         - name: b
           datatype: ['ascii', 16]
    """
    buff = helpers.yaml_to_asdf(content)

    with asdf.open(buff):
        pass


def test_string_inline():
    x = np.array([b"a", b"b", b"c"])
    line = ndarray.numpy_array_to_list(x)

    for entry in line:
        assert isinstance(entry, str)


def test_inline_shape_mismatch(ndarray_tag):
    content = f"""
arr: !{ndarray_tag}
  data: [1, 2, 3]
  shape: [2]
    """

    buff = helpers.yaml_to_asdf(content)
    with pytest.raises(ValueError, match=r"inline data doesn't match the given shape"):
        with asdf.open(buff) as af:
            af["arr"]


def test_broadcasted_array():
    attrs = np.broadcast_arrays(np.array([10, 20]), np.array(10), np.array(10))
    tree = {"one": attrs[1]}  # , 'two': attrs[1], 'three': attrs[2]}
    with roundtrip(tree) as af:
        assert_array_equal(tree["one"], af["one"])


def test_broadcasted_offset_array():
    base = np.arange(10)
    offset = base[5:]
    broadcasted = np.broadcast_to(offset, (4, 5))
    tree = {"broadcasted": broadcasted}
    with roundtrip(tree) as af:
        assert_array_equal(tree["broadcasted"], af["broadcasted"])


def test_non_contiguous_base_array():
    base = np.arange(60).reshape(5, 4, 3).transpose(2, 0, 1) * 1
    contiguous = base.transpose(1, 2, 0)
    tree = {"contiguous": contiguous}
    with roundtrip(tree) as af:
        assert_array_equal(tree["contiguous"], af["contiguous"])


def test_fortran_order():
    array = np.array([[11, 12, 13], [21, 22, 23]], order="F", dtype=np.int64)
    tree = {"data": array}

    with roundtrip(tree) as af:
        assert af["data"].flags.fortran
        assert np.all(np.isclose(array, af["data"]))

    with roundtrip(tree, raw=True) as content:
        tree = yaml.safe_load(re.sub(rb"!core/\S+", b"", content))
        assert tree["data"]["strides"] == [8, 16]


def test_memmap_write(tmp_path):
    tmpfile = str(tmp_path / "data.asdf")
    tree = {"data": np.zeros(100)}

    with asdf.AsdfFile(tree, memmap=False) as af:
        # Make sure we're actually writing to an internal array for this test
        af.write_to(tmpfile, all_array_storage="internal")

    with asdf.open(tmpfile, mode="rw", memmap=True) as af:
        data = af["data"]
        assert data.flags.writeable is True
        data[0] = 42
        assert data[0] == 42

    with asdf.open(tmpfile, mode="rw", memmap=True) as af:
        assert af["data"][0] == 42

    with asdf.open(tmpfile, mode="r", memmap=True) as af:
        assert af["data"][0] == 42


def test_readonly(tmp_path):
    tmpfile = str(tmp_path / "data.asdf")
    tree = {"data": np.ndarray(100)}

    with asdf.AsdfFile(tree) as af:
        # Make sure we're actually writing to an internal array for this test
        af.write_to(tmpfile, all_array_storage="internal")

    # Opening in read mode (the default) should mean array is readonly
    with asdf.open(tmpfile, memmap=True) as af:
        assert af["data"].flags.writeable is False
        with pytest.raises(ValueError, match=r"assignment destination is read-only"):
            af["data"][0] = 41

    # Forcing memmap, the array should still be readonly
    with asdf.open(tmpfile, memmap=True) as af:
        assert af["data"].flags.writeable is False
        with pytest.raises(ValueError, match=r"assignment destination is read-only"):
            af["data"][0] = 41

    # This should be perfectly fine
    with asdf.open(tmpfile, mode="rw") as af:
        assert af["data"].flags.writeable is True
        af["data"][0] = 40

    # Copying the arrays makes it safe to write to the underlying array
    with asdf.open(tmpfile, mode="r", memmap=False) as af:
        assert af["data"].flags.writeable is True
        af["data"][0] = 42


def test_readonly_inline(tmp_path):
    tmpfile = str(tmp_path / "data.asdf")
    tree = {"data": np.ndarray(100)}

    with asdf.AsdfFile(tree) as af:
        af.write_to(tmpfile, all_array_storage="inline")

    # This should be safe since it's an inline array
    with asdf.open(tmpfile, mode="r") as af:
        assert af["data"].flags.writeable is True
        af["data"][0] = 42


# Confirm that NDArrayType's internal array is regenerated
# following an update.
@pytest.mark.parametrize("pad_blocks", [True, False])
def test_block_data_change(pad_blocks, tmp_path):
    tmpfile = str(tmp_path / "data.asdf")
    tree = {"data": np.zeros(10, dtype="uint8")}
    with asdf.AsdfFile(tree) as af:
        af.write_to(tmpfile, pad_blocks=pad_blocks)

    with asdf.open(tmpfile, mode="rw", memmap=True) as af:
        assert np.all(af.tree["data"] == 0)
        array_before = af.tree["data"].__array__()
        af.tree["data"][:5] = 1
        af.update()
        array_after = af.tree["data"].__array__()
        assert array_before is not array_after
        assert np.all(af.tree["data"][:5] == 1)
        assert np.all(af.tree["data"][5:] == 0)


def test_problematic_class_attributes(tmp_path):
    """
    The presence of the "name" and "version" attributes
    in NDArrayType cause problems when our arrays are used
    with other libraries.

    See https://github.com/asdf-format/asdf/issues/1015
    """
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["arr"] = np.arange(100)
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert isinstance(af["arr"], ndarray.NDArrayType)

        with pytest.raises(AttributeError, match=r".* object has no attribute 'name'"):
            af["arr"].name

        with pytest.raises(AttributeError, match=r".* object has no attribute 'version'"):
            af["arr"].version


def test_shape_does_not_load_array(tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["arr"] = np.arange(100)
        af.write_to(file_path)

    with asdf.open(file_path, lazy_load=True) as af:
        assert af["arr"]._array is None
        assert af["arr"].shape == (100,)
        assert af["arr"]._array is None


@pytest.mark.parametrize(
    "lazy_load, array_class",
    (
        (True, ndarray.NDArrayType),
        (False, np.ndarray),
    ),
)
@pytest.mark.parametrize("lazy_tree", [True, False])
def test_lazy_load_array_class(tmp_path, lazy_load, lazy_tree, array_class):
    file_path = tmp_path / "test.asdf"
    asdf.AsdfFile({"arr": np.arange(100)}).write_to(file_path)

    with asdf.open(file_path, lazy_load=lazy_load, lazy_tree=lazy_tree) as af:
        assert type(af["arr"]) is array_class



from .key import Key


class Store:
    """
    A key-value store that uses ``asdf._block.key.Key``
    to allow use of keys that:
        - are not hashable (so any object can be used)
        - when the key is garbage collected, the value
          will be unretrievable
    """

    def __init__(self):
        # store contains 2 layers of lookup: id(obj), Key
        self._by_id = {}

    def lookup_by_object(self, obj, default=None):
        if isinstance(obj, Key):
            # if obj is a Key, look up the object
            obj_id = id(obj._ref())
            # and use the Key
            obj_key = obj
        else:
            obj_id = id(obj)
            obj_key = None

        # if id is unknown, return default
        if obj_id not in self._by_id:
            return default

        # first, lookup by id: O(1)
        by_key = self._by_id[obj_id]

        # if we have a key
        if obj_key:
            # use the key to get an existing value
            # or default if this Key is unknown
            return by_key.get(obj_key, default)

        # we have seen this id(obj) before
        # look for a matching key: O(N)
        for key, value in by_key.items():
            if key._matches_object(obj):
                return value

        # no match, return default
        return default

    def assign_object(self, obj, value):
        if isinstance(obj, Key):
            if not obj._is_valid():
                msg = "Invalid key used for assign_object"
                raise ValueError(msg)
            obj_id = id(obj._ref())
            obj_key = obj
        else:
            obj_id = id(obj)
            obj_key = None

        # if the id is unknown, just set it
        if obj_id not in self._by_id:
            if obj_key is None:
                obj_key = Key(obj)
            self._by_id[obj_id] = {obj_key: value}
            return

        # if id is known
        by_key = self._by_id[obj_id]

        # look for a matching matching key
        if obj_key is None:
            for key in by_key:
                if key._matches_object(obj):
                    by_key[key] = value
                    return
            # we didn't find a matching key, so make one
            obj_key = Key(obj)

        # if no match was found, add using the key
        self._by_id[obj_id][obj_key] = value

    def keys_for_value(self, value):
        for oid, by_key in self._by_id.items():
            for key, stored_value in by_key.items():
                if stored_value == value and key._is_valid():
                    yield key

    def _cleanup(self, object_id=None):
        if object_id is None:
            for oid in set(self._by_id):
                self._cleanup(oid)
            return
        by_key = self._by_id[object_id]
        keys_to_remove = [k for k in by_key if not k._is_valid()]
        for key in keys_to_remove:
            del by_key[key]
        if not len(by_key):
            del self._by_id[object_id]



import numpy as np
import pytest

import asdf
from asdf._commands import main


@pytest.fixture()
def valid_file_path(tmp_path):
    path = tmp_path / "valid_file.asdf"
    asdf.dump({"foo": 42, "arr": np.arange(42)}, path)
    return path


@pytest.fixture()
def invalid_file_path(tmp_path):
    # don't use asdf as we're intentionally making an invalid file
    contents = b"""#ASDF 1.0.0
#ASDF_STANDARD 1.6.0
%YAML 1.1
%TAG ! tag:stsci.edu:asdf/
--- !core/asdf-1.1.0
arr: !core/ndarray-1.1.0
  data:
    nested: error
..."""
    path = tmp_path / "invalid_file.asdf"
    with path.open("wb") as f:
        f.write(contents)
    return path


@pytest.fixture()
def custom_schema_path(tmp_path):
    contents = """%YAML 1.1
---
id: "http://example.com/schemas/your-custom-schema"
$schema: "http://stsci.edu/schemas/yaml-schema/draft-01"
type: object
properties:
  foo:
    type: string"""
    path = tmp_path / "custom_schema.yaml"
    with path.open("w") as f:
        f.write(contents)
    return path


@pytest.fixture()
def bad_blocks_file_path(tmp_path):
    to_replace = b"REPLACE"
    replacement = b"ABCDEFG"
    buff = asdf.dumps({"arr": np.frombuffer(to_replace, np.uint8)})
    buff = buff.replace(to_replace, replacement)
    path = tmp_path / "bad_block.asdf"
    with path.open("wb") as f:
        f.write(buff)
    return path


def test_valid(capsys, valid_file_path):
    assert main.main_from_args(["validate", str(valid_file_path)]) == 0

    captured = capsys.readouterr()
    assert "valid" in captured.out


def test_invalid(invalid_file_path):
    with pytest.raises(asdf.ValidationError):
        main.main_from_args(["validate", str(invalid_file_path)])


def test_custom_schema(valid_file_path, custom_schema_path):
    with pytest.raises(asdf.ValidationError):
        main.main_from_args(["validate", str(valid_file_path), "--custom-schema", str(custom_schema_path)])


def test_block_checksum(bad_blocks_file_path):
    with pytest.raises(ValueError, match="does not match given checksum"):
        main.main_from_args(["validate", str(bad_blocks_file_path)])


def test_skip_block_validation(bad_blocks_file_path):
    assert main.main_from_args(["validate", str(bad_blocks_file_path), "--skip-block-validation"]) == 0



import pytest

import asdf
from asdf.exceptions import AsdfConversionWarning
from asdf.testing.helpers import yaml_to_asdf


def test_undefined_tag(with_lazy_tree):
    # This tests makes sure that ASDF still returns meaningful structured data
    # even when it encounters a schema tag that it does not specifically
    # implement as an extension
    from numpy import array

    undefined_tag = "tag:nowhere.org:custom/also_undefined-1.3.0"
    yaml = f"""
undefined_data:
  !<tag:nowhere.org:custom/undefined_tag-1.0.0>
    - 5
    - 'message': 'there is no tag'
    - !core/ndarray-1.1.0
      [[1, 2, 3], [4, 5, 6]]
    - !<{undefined_tag}>
        - !core/ndarray-1.1.0 [[7],[8],[9],[10]]
        - !core/complex-1.0.0 3.14j
"""
    buff = yaml_to_asdf(yaml, version="1.6.0")
    with pytest.warns(Warning) as warning:
        afile = asdf.open(buff)
        missing = afile.tree["undefined_data"]
        missing[3]

    assert missing[0] == 5
    assert missing[1] == {"message": "there is no tag"}
    assert isinstance(missing[3], asdf.tagged.TaggedList)
    assert missing[3]._tag == undefined_tag
    if with_lazy_tree:
        assert isinstance(missing[3].data, asdf.lazy_nodes.AsdfListNode)
    assert (missing[2] == array([[1, 2, 3], [4, 5, 6]])).all()
    assert (missing[3][0] == array([[7], [8], [9], [10]])).all()
    assert missing[3][1] == 3.14j

    # There are two undefined tags, so we expect two warnings
    # filter out only AsdfConversionWarning
    warning = [w for w in warning if w.category == AsdfConversionWarning]
    assert len(warning) == 2
    messages = {str(w.message) for w in warning}
    match = {
        f"tag:nowhere.org:custom/{tag} is not recognized, converting to raw Python data structure"
        for tag in ("undefined_tag-1.0.0", "also_undefined-1.3.0")
    }
    assert messages == match

    # Make sure no warning occurs if explicitly ignored
    buff.seek(0)
    # as warnings get turned into errors, nothing to do here
    afile = asdf.open(buff, ignore_unrecognized_tag=True)



"""
Utility functions for managing tree-like data structures.
"""

import collections
import types
from contextlib import contextmanager

from . import lazy_nodes, tagged

__all__ = ["PendingValue", "RemoveNode", "get_children", "is_container", "iter_tree", "walk", "walk_and_modify"]


def walk(top, callback):
    """
    Walking through a tree of objects, calling a given function at
    each node.

    Parameters
    ----------
    top : object
        The root of the tree.  May be a dict, list or other Python object.

    callback : callable
        A function to call at each node in the tree.

        The callback is called on an instance after all of its
        children have been visited (depth-first order).

    Returns
    -------
    tree : object
        The modified tree.
    """
    for x in iter_tree(top):
        callback(x)


def iter_tree(top):
    """
    Iterate over all nodes in a tree, in depth-first order.

    Parameters
    ----------
    top : object
        The root of the tree.  May be a dict, list or other Python object.

    callback : callable
        A function to call at each node in the tree.

        The callback is called on an instance after all of its
        children have been visited (depth-first order).

    Returns
    -------
    tree : object
        The modified tree.
    """
    seen = set()

    def recurse(tree):
        tree_id = id(tree)

        if tree_id in seen:
            return

        if isinstance(tree, (list, tuple, lazy_nodes.AsdfListNode)):
            seen.add(tree_id)
            for val in tree:
                yield from recurse(val)
            seen.remove(tree_id)
        elif isinstance(tree, (dict, lazy_nodes.AsdfDictNode)):
            seen.add(tree_id)
            for val in tree.values():
                yield from recurse(val)
            seen.remove(tree_id)

        yield tree

    return recurse(top)


class _TreeModificationContext:
    """
    Context of a call to walk_and_modify, which includes a map
    of already modified nodes, a list of generators to drain
    before exiting the call, and a set of node object ids that
    are currently pending modification.

    Instances of this class are context managers that track
    how many times they have been entered, and only drain
    generators and reset themselves when exiting the outermost
    context.  They are also collections that map unmodified
    nodes to the corresponding modified result.
    """

    def __init__(self):
        self._map = {}
        self._generators = []
        self._depth = 0
        self._pending = set()

    def add_generator(self, generator):
        """
        Add a generator that should be drained before exiting
        the outermost call to walk_and_modify.
        """
        self._generators.append(generator)

    def is_pending(self, node):
        """
        Return True if the node is already being modified.
        This will not be the case unless the node contains a
        reference to itself somewhere among its descendents.
        """
        return id(node) in self._pending

    @contextmanager
    def pending(self, node):
        """
        Context manager that marks a node as pending for the
        duration of the context.
        """
        if id(node) in self._pending:
            msg = (
                "Unhandled cycle in tree.  This is possibly a bug "
                "in extension code, which should be yielding "
                "nodes that may contain reference cycles."
            )
            raise RuntimeError(msg)

        self._pending.add(id(node))
        try:
            yield self
        finally:
            self._pending.remove(id(node))

    def __enter__(self):
        self._depth += 1
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._depth -= 1

        if self._depth == 0:
            # If we're back to 0 depth, then we're exiting
            # the outermost context, so it's time to drain
            # the generators and reset this object for next
            # time.
            if exc_type is None:
                self._drain_generators()
            self._generators = []
            self._map = {}
            self._pending = set()

    def _drain_generators(self):
        """
        Drain each generator we've accumulated during this
        call to walk_and_modify.
        """
        # Generator code may add yet more generators
        # to the list, so we need to loop until the
        # list is empty.
        while len(self._generators) > 0:
            generators = self._generators
            self._generators = []
            for generator in generators:
                for _ in generator:
                    # Subsequent yields of the generator should
                    # always return the same value.  What we're
                    # really doing here is executing the generator's
                    # remaining code, to further modify that first
                    # yielded object.
                    pass

    def __contains__(self, node):
        return id(node) in self._map

    def __getitem__(self, node):
        return self._map[id(node)][1]

    def __setitem__(self, node, result):
        if id(node) in self._map:
            # This indicates that an already defined
            # modified node is being replaced, which is an
            # error because it breaks references within the
            # tree.
            msg = "Node already has an associated result"
            raise RuntimeError(msg)

        self._map[id(node)] = (node, result)


class _PendingValue:
    """
    Class of the PendingValue singleton instance.  The presence of the instance
    in an asdf tree indicates that extension code is failing to handle
    reference cycles.
    """

    def __repr__(self):
        return "PendingValue"


PendingValue = _PendingValue()


class _RemoveNode:
    """
    Class of the RemoveNode singleton instance.  This instance is used
    as a signal for `asdf.treeutil.walk_and_modify` to remove the
    node received by the callback.
    """

    def __repr__(self):
        return "RemoveNode"


RemoveNode = _RemoveNode()


def walk_and_modify(top, callback, postorder=True, _context=None):
    """Modify a tree by walking it with a callback function.  It also has
    the effect of doing a deep copy.

    Parameters
    ----------
    top : object
        The root of the tree.  May be a dict, list or other Python object.

    callback : callable
        A function to call at each node in the tree.  It takes either
        one or two arguments:

        - an instance from the tree
        - a json id (optional)

        It may return a different instance in order to modify the
        tree.  If the singleton instance `~asdf.treeutil._RemoveNode`
        is returned, the node will be removed from the tree.

        The json id is the context under which any relative URLs
        should be resolved.  It may be `None` if no ids are in the file

        The tree is traversed depth-first, with order specified by the
        ``postorder`` argument.

    postorder : bool
        Determines the order in which the callable is invoked on nodes of
        the tree.  If `True`, the callable will be invoked on children
        before their parents.  If `False`, the callable is invoked on the
        parents first.  Defaults to `True`.

    Returns
    -------
    tree : object
        The modified tree.

    """
    callback_arity = callback.__code__.co_argcount
    if callback_arity < 1 or callback_arity > 2:
        msg = "Expected callback to accept one or two arguments"
        raise ValueError(msg)

    def _handle_generator(result):
        # If the result is a generator, generate one value to
        # extract the true result, then register the generator
        # to be drained later.
        if isinstance(result, types.GeneratorType):
            generator = result
            result = next(generator)
            _context.add_generator(generator)

        return result

    def _handle_callback(node, json_id):
        result = callback(node) if callback_arity == 1 else callback(node, json_id)

        return _handle_generator(result)

    def _handle_mapping(node, json_id):
        if isinstance(node, lazy_nodes.AsdfOrderedDictNode):
            result = collections.OrderedDict()
        elif isinstance(node, lazy_nodes.AsdfDictNode):
            result = {}
        else:
            result = node.__class__()
        if isinstance(node, tagged.Tagged):
            result._tag = node._tag

        pending_items = {}
        for key, value in node.items():
            if _context.is_pending(value):
                # The child node is pending modification, which means
                # it must be its own ancestor.  Assign the special
                # PendingValue instance for now, and note that we'll
                # need to fill in the real value later.
                pending_items[key] = value
                result[key] = PendingValue

            elif (val := _recurse(value, json_id)) is not RemoveNode:
                result[key] = val

        yield result

        if len(pending_items) > 0:
            # Now that we've yielded, the pending children should
            # be available.
            for key, value in pending_items.items():
                if (val := _recurse(value, json_id)) is not RemoveNode:
                    result[key] = val
                else:
                    # The callback may have decided to delete
                    # this node after all.
                    del result[key]

    def _handle_mutable_sequence(node, json_id):
        if isinstance(node, lazy_nodes.AsdfListNode):
            result = []
        else:
            result = node.__class__()
        if isinstance(node, tagged.Tagged):
            result._tag = node._tag

        pending_items = {}
        for i, value in enumerate(node):
            if _context.is_pending(value):
                # The child node is pending modification, which means
                # it must be its own ancestor.  Assign the special
                # PendingValue instance for now, and note that we'll
                # need to fill in the real value later.
                pending_items[i] = value
                result.append(PendingValue)
            else:
                result.append(_recurse(value, json_id))

        yield result

        for i, value in pending_items.items():
            # Now that we've yielded, the pending children should
            # be available.
            result[i] = _recurse(value, json_id)

    def _handle_immutable_sequence(node, json_id):
        # Immutable sequences containing themselves are impossible
        # to construct (well, maybe possible in a C extension, but
        # we're not going to worry about that), so we don't need
        # to yield here.
        contents = [_recurse(value, json_id) for value in node]

        result = node.__class__(contents)
        if isinstance(node, tagged.Tagged):
            result._tag = node._tag

        return result

    def _handle_children(node, json_id):
        if isinstance(node, (dict, lazy_nodes.AsdfDictNode)):
            result = _handle_mapping(node, json_id)
        # don't treat namedtuple instances as tuples
        # see: https://github.com/python/cpython/issues/52044
        elif isinstance(node, tuple) and not hasattr(node, "_fields"):
            result = _handle_immutable_sequence(node, json_id)
        elif isinstance(node, (list, lazy_nodes.AsdfListNode)):
            result = _handle_mutable_sequence(node, json_id)
        else:
            result = node

        return _handle_generator(result)

    def _recurse(node, json_id=None):
        if node in _context:
            # The node's modified result has already been
            # created, all we need to do is return it.  This
            # occurs when the tree contains multiple references
            # to the same object id.
            return _context[node]

        # Inform the context that we're going to start modifying
        # this node.
        with _context.pending(node):
            # Take note of the "id" field, in case we're modifying
            # a schema and need to know the namespace for resolving
            # URIs.  Ignore an id that is not a string, since it may
            # be an object defining an id property and not an id
            # itself (this is common in metaschemas).
            if isinstance(node, (dict, lazy_nodes.AsdfDictNode)) and "id" in node and isinstance(node["id"], str):
                json_id = node["id"]

            if postorder:
                # If this is a postorder modification, invoke the
                # callback on this node's children first.
                result = _handle_children(node, json_id)
                result = _handle_callback(result, json_id)
            else:
                # Otherwise, invoke the callback on the node first,
                # then its children.
                result = _handle_callback(node, json_id)
                result = _handle_children(result, json_id)

        # Store the result in the context, in case there are
        # additional references to the same node elsewhere in
        # the tree.
        _context[node] = result

        return result

    if _context is None:
        _context = _TreeModificationContext()

    with _context:
        return _recurse(top)
        # Generators will be drained here, if this is the outermost
        # call to walk_and_modify.


def get_children(node):
    """
    Retrieve the children (and their dict keys or list/tuple indices) of
    an ASDF tree node.

    Parameters
    ----------
    node : object
        an ASDF tree node

    Returns
    -------
    list of (object, object) tuples
        list of (identifier, child node) tuples, or empty list if the
        node has no children (either it is an empty container, or is
        a non-container type)
    """
    if isinstance(node, (dict, lazy_nodes.AsdfDictNode)):
        return list(node.items())

    if isinstance(node, (list, tuple, lazy_nodes.AsdfListNode)):
        return list(enumerate(node))

    return []


def is_container(node):
    """
    Determine if an ASDF tree node is an instance of a "container" type
    (i.e., value may contain child nodes).

    Parameters
    ----------
    node : object
        an ASDF tree node

    Returns
    -------
    bool
        True if node is a container, False otherwise
    """
    return isinstance(node, (dict, list, tuple, lazy_nodes.AsdfDictNode, lazy_nodes.AsdfListNode))



"""
Integration tests for the new plugin APIs.
"""

import pytest

import asdf
from asdf.extension import TagDefinition

FOO_SCHEMA_URI = "asdf://somewhere.org/extensions/foo/schemas/foo-1.0"
FOO_SCHEMA = f"""
id: {FOO_SCHEMA_URI}
type: object
properties:
  value:
    type: string
required: ["value"]
"""


class Foo:
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value


class FooConverter:
    types = [Foo]
    tags = ["asdf://somewhere.org/extensions/foo/tags/foo-*"]

    def to_yaml_tree(self, obj, tag, ctx):
        return {"value": obj.value}

    def from_yaml_tree(self, obj, tag, ctx):
        return Foo(obj["value"])


class FooExtension:
    extension_uri = "asdf://somewhere.org/extensions/foo-1.0"
    converters = [FooConverter()]
    tags = [
        TagDefinition(
            "asdf://somewhere.org/extensions/foo/tags/foo-1.0",
            schema_uris=FOO_SCHEMA_URI,
        ),
    ]


def test_serialize_custom_type(tmp_path):
    with asdf.config_context() as config:
        config.add_resource_mapping({FOO_SCHEMA_URI: FOO_SCHEMA})
        config.add_extension(FooExtension())

        path = str(tmp_path / "test.asdf")

        af = asdf.AsdfFile()
        af["foo"] = Foo("bar")
        af.write_to(path)

        with asdf.open(path) as af2:
            assert af2["foo"].value == "bar"

        af["foo"] = Foo(12)
        with pytest.raises(asdf.ValidationError, match=r".* is not of type .*"):
            af.write_to(path)


FOOFOO_SCHEMA_URI = "asdf://somewhere.org/extensions/foo/schemas/foo_foo-1.0"
FOOFOO_SCHEMA = f"""
id: {FOOFOO_SCHEMA_URI}
type: object
properties:
  value_value:
    type: string
required: ["value_value"]
"""


class FooFoo(Foo):
    def __init__(self, value, value_value):
        super().__init__(value)

        self._value_value = value_value

    @property
    def value_value(self):
        return self._value_value


class FooFooConverter:
    types = [FooFoo]
    tags = ["asdf://somewhere.org/extensions/foo/tags/foo_foo-*"]

    def to_yaml_tree(self, obj, tag, ctx):
        return {"value": obj.value, "value_value": obj.value_value}

    def from_yaml_tree(self, obj, tag, ctx):
        return FooFoo(obj["value"], obj["value_value"])


class FooFooExtension:
    extension_uri = "asdf://somewhere.org/extensions/foo_foo-1.0"
    converters = [FooFooConverter()]
    tags = [
        TagDefinition(
            "asdf://somewhere.org/extensions/foo/tags/foo_foo-1.0",
            schema_uris=[FOO_SCHEMA_URI, FOOFOO_SCHEMA_URI],
        ),
    ]


def test_serialize_with_multiple_schemas(tmp_path):
    with asdf.config_context() as config:
        config.add_resource_mapping({FOO_SCHEMA_URI: FOO_SCHEMA, FOOFOO_SCHEMA_URI: FOOFOO_SCHEMA})
        config.add_extension(FooFooExtension())

        path = str(tmp_path / "test.asdf")

        af = asdf.AsdfFile()
        af["foo_foo"] = FooFoo("bar", "bar_bar")
        af.write_to(path)

        with asdf.open(path) as af2:
            assert af2["foo_foo"].value == "bar"
            assert af2["foo_foo"].value_value == "bar_bar"

        af["foo_foo"] = FooFoo(12, "bar_bar")
        with pytest.raises(asdf.ValidationError, match=r".* is not of type .*"):
            af.write_to(path)

        af["foo_foo"] = FooFoo("bar", 34)
        with pytest.raises(asdf.ValidationError, match=r".* is not of type .*"):
            af.write_to(path)


class FooFooConverterlessExtension:
    extension_uri = "asdf://somewhere.org/extensions/foo_foo-1.0"
    converters = []
    tags = [
        TagDefinition(
            "asdf://somewhere.org/extensions/foo/tags/foo_foo-1.0",
            schema_uris=[FOO_SCHEMA_URI, FOOFOO_SCHEMA_URI],
        ),
    ]


def test_converterless_serialize_with_multiple_schemas(tmp_path):
    with asdf.config_context() as config:
        config.add_resource_mapping({FOO_SCHEMA_URI: FOO_SCHEMA, FOOFOO_SCHEMA_URI: FOOFOO_SCHEMA})
        config.add_extension(FooFooConverterlessExtension())

        path = str(tmp_path / "test.asdf")

        af = asdf.AsdfFile()
        af["foo_foo"] = "asdf://somewhere.org/extensions/foo/tags/foo_foo-1.0 {bar: bar_bar}"
        af.write_to(path)

        with asdf.open(path) as af2:
            assert af2["foo_foo"] == af["foo_foo"]



"""
Helpers for writing unit tests of ASDF support.
"""

from io import BytesIO

import asdf


def roundtrip_object(obj, version=None):
    """
    Add the specified object to an AsdfFile's tree, write the file to
    a buffer, then read it back in and return the deserialized object.

    Parameters
    ----------
    obj : object
        Object to serialize.
    version : str or None.
        ASDF Standard version.  If None, use the library's default version.

    Returns
    -------
    object
        The deserialized object.
    """
    buff = BytesIO()
    with asdf.AsdfFile(version=version) as af:
        af["obj"] = obj
        af.write_to(buff)

    buff.seek(0)
    with asdf.open(buff, lazy_load=False, memmap=False) as af:
        return af["obj"]


def yaml_to_asdf(yaml_content, version=None):
    """
    Given a string of YAML content, adds the extra pre-
    and post-amble to make it an ASDF file.

    Parameters
    ----------
    yaml_content : string or bytes
        YAML content.
    version : str or None.
        ASDF Standard version.  If None, use the library's default version.

    Returns
    -------
    io.BytesIO
        A file-like object containing the ASDF file.
    """
    if isinstance(yaml_content, str):
        yaml_content = yaml_content.encode("utf-8")

    buff = BytesIO()
    with asdf.AsdfFile(version=version) as af:
        af["$REPLACE"] = "ME"
        af.write_to(buff)

    buff.seek(0)
    asdf_content = buff.read().replace(b"$REPLACE: ME", yaml_content)

    return BytesIO(asdf_content)



class ExternalArrayReference:
    """
    Store a reference to an array in an external File.

    This class is a simple way of referring to an array in another file. It
    provides no way to resolve these references, that is left to the user. It
    also performs no checking to see if any of the arguments are correct. e.g.
    if the file exits.

    Parameters
    ----------

    fileuri: `str`
        The path to the path to be referenced. Can be relative to the file
        containing the reference.

    target: `object`
        Some internal target to the data in the file. Examples may include a HDU
        index, a HDF path or an asdf fragment.

    dtype: `str`
        The (numpy) dtype of the contained array.

    shape: `tuple`
        The shape of the array to be loaded.


    Examples
    --------

    >>> import asdf
    >>> ref = asdf.ExternalArrayReference("myfitsfile.fits", 1, "float64", (100, 100))
    >>> tree = {'reference': ref}
    >>> with asdf.AsdfFile(tree) as ff:
    ...     ff.write_to("test.asdf")

    """

    def __init__(self, fileuri, target, dtype, shape):
        self.fileuri = str(fileuri)
        self.target = target
        self.dtype = dtype
        self.shape = tuple(shape)

    def __repr__(self):
        return f"<External array reference in {self.fileuri} at {self.target} shape: {self.shape} dtype: {self.dtype}>"

    def __str__(self):
        return repr(self)

    def __eq__(self, other):
        uri = self.fileuri == other.fileuri
        target = self.target == other.target
        dtype = self.dtype == other.dtype
        shape = self.shape == other.shape

        return all((uri, target, dtype, shape))



import io
import os
import re
import stat
import sys
import urllib.request as urllib_request

import numpy as np
import pytest

import asdf
from asdf import exceptions, generic_io
from asdf.config import config_context

from . import _helpers as helpers


def _roundtrip(tree, get_write_fd, get_read_fd, write_options=None, read_options=None):
    write_options = {} if write_options is None else write_options
    read_options = {} if read_options is None else read_options

    # Since we're testing with small arrays, force all arrays to be stored
    # in internal blocks rather than letting some of them be automatically put
    # inline.
    write_options.setdefault("all_array_storage", "internal")

    with get_write_fd() as fd:
        asdf.AsdfFile(tree).write_to(fd, **write_options)
        # Work around the fact that generic_io's get_file doesn't have a way of
        # determining whether or not the underlying file handle should be
        # closed as part of the exit handler
        if isinstance(fd._fd, io.FileIO):
            fd._fd.close()

    with get_read_fd() as fd:
        ff = asdf.open(fd, **read_options)
        helpers.assert_tree_match(tree, ff.tree)

    return ff


def test_mode_fail(tmp_path):
    path = os.path.join(str(tmp_path), "test.asdf")

    with pytest.raises(ValueError, match=r"mode must be 'r', 'w' or 'rw'"):
        generic_io.get_file(path, mode="r+")


@pytest.mark.parametrize("mode", ["r", "w", "rw"])
def test_missing_directory(tmp_path, mode):
    path = str(tmp_path / "missing" / "test.asdf")
    regex_path = re.escape(path.replace("\\", "\\\\"))
    with pytest.raises(FileNotFoundError, match=f".*: '{regex_path}'$"):
        generic_io.get_file(path, mode=mode)


@pytest.mark.skipif(sys.platform.startswith("win"), reason="paths with two leading slashes are valid on windows")
@pytest.mark.parametrize("mode", ["r", "w", "rw"])
def test_two_leading_slashes(mode):
    """
    Providing a path with two leading slashes '//' will be parsed
    by urllib as having a netloc (unhandled by generic_io) and
    an invalid path. This creates an unhelpful error message on
    write (related to a missing atomic write file) and should be
    cause beforehand and provided with a more helpful error message

    Regression test for issue:
    https://github.com/asdf-format/asdf/issues/1353
    """
    path = "//bad/two/slashes"
    with pytest.raises(ValueError, match="Invalid path"):
        generic_io.get_file(path, mode=mode)


def test_open(tmp_path, small_tree):
    from asdf import open

    path = os.path.join(str(tmp_path), "test.asdf")

    # Simply tests the high-level "open" function
    ff = asdf.AsdfFile(small_tree)
    ff.write_to(path)
    with open(path) as ff2:
        helpers.assert_tree_match(ff2.tree, ff.tree)


def test_path(tree, tmp_path):
    path = tmp_path / "test.asdf"

    def get_write_fd():
        f = generic_io.get_file(path, mode="w")
        assert isinstance(f, generic_io.RealFile)
        assert f._uri == path.as_uri()
        return f

    def get_read_fd():
        # Must open with mode=rw in order to get memmapped data
        f = generic_io.get_file(path, mode="rw")
        assert isinstance(f, generic_io.RealFile)
        assert f._uri == path.as_uri()
        # This is to check for a "feature" in Python 3.x that reading zero
        # bytes from a socket causes it to stop.  We have code in generic_io.py
        # to workaround it.
        f.read(0)
        return f

    with _roundtrip(tree, get_write_fd, get_read_fd, read_options={"memmap": True}) as ff:
        assert len(ff._blocks.blocks) == 2
        assert isinstance(ff._blocks.blocks[0].cached_data, np.memmap)


def test_open2(tree, tmp_path):
    path = tmp_path / "test.asdf"

    def get_write_fd():
        # cannot use context manager here because it closes the file
        f = generic_io.get_file(open(path, "wb"), mode="w", close=True)
        assert isinstance(f, generic_io.RealFile)
        assert f._uri == path.as_uri()
        return f

    def get_read_fd():
        # Must open with mode=rw in order to get memmapped data
        # cannot use context manager here because it closes the file
        f = generic_io.get_file(open(path, "r+b"), mode="rw", close=True)
        assert isinstance(f, generic_io.RealFile)
        assert f._uri == path.as_uri()
        return f

    with _roundtrip(tree, get_write_fd, get_read_fd, read_options={"memmap": True}) as ff:
        assert len(ff._blocks.blocks) == 2
        assert isinstance(ff._blocks.blocks[0].cached_data, np.memmap)


@pytest.mark.parametrize("mode", ["r", "w", "rw"])
def test_open_not_binary_fail(tmp_path, mode):
    path = tmp_path / "test.asdf"

    with open(path, "w") as fd:
        fd.write("\n\n\n")

    file_mode = mode if mode != "rw" else "r+"
    with (
        open(path, file_mode) as fd,
        pytest.raises(
            ValueError,
            match=r"File-like object must be opened in binary mode.",
        ),
    ):
        generic_io.get_file(fd, mode=mode)


def test_io_open(tree, tmp_path):
    path = tmp_path / "test.asdf"

    def get_write_fd():
        # cannot use context manager here because it closes the file
        f = generic_io.get_file(open(path, "wb"), mode="w", close=True)
        assert isinstance(f, generic_io.RealFile)
        assert f._uri == path.as_uri()
        return f

    def get_read_fd():
        # cannot use context manager here because it closes the file
        f = generic_io.get_file(open(path, "r+b"), mode="rw", close=True)
        assert isinstance(f, generic_io.RealFile)
        assert f._uri == path.as_uri()
        return f

    with _roundtrip(tree, get_write_fd, get_read_fd, read_options={"memmap": True}) as ff:
        assert len(ff._blocks.blocks) == 2
        assert isinstance(ff._blocks.blocks[0].cached_data, np.memmap)
        ff.tree["science_data"][0] = 42


def test_close_underlying(tmp_path):
    path = os.path.join(str(tmp_path), "test.asdf")

    with generic_io.get_file(open(path, "wb"), mode="w", close=True) as ff:
        pass

    assert ff.is_closed() is True
    assert ff._fd.closed is True

    with generic_io.get_file(open(path, "rb"), close=True) as ff2:
        pass

    assert ff2.is_closed() is True
    assert ff2._fd.closed is True


def test_bytes_io(tree):
    buff = io.BytesIO()

    def get_write_fd():
        f = generic_io.get_file(buff, mode="w")
        assert isinstance(f, generic_io.MemoryIO)
        return f

    def get_read_fd():
        buff.seek(0)
        f = generic_io.get_file(buff, mode="rw")
        assert isinstance(f, generic_io.MemoryIO)
        return f

    with _roundtrip(tree, get_write_fd, get_read_fd) as ff:
        assert len(ff._blocks.blocks) == 2
        assert not isinstance(ff._blocks.blocks[0].cached_data, np.memmap)
        ff.tree["science_data"][0] = 42


def test_streams(tree):
    buff = io.BytesIO()

    def get_write_fd():
        return generic_io.OutputStream(buff)

    def get_read_fd():
        buff.seek(0)
        return generic_io.InputStream(buff, "rw")

    with _roundtrip(tree, get_write_fd, get_read_fd) as ff:
        assert len(ff._blocks.blocks) == 2
        assert not isinstance(ff._blocks.blocks[0].cached_data, np.memmap)
        ff.tree["science_data"][0] = 42


def test_streams2():
    buff = io.BytesIO(b"\0" * 60)
    buff.seek(0)

    fd = generic_io.InputStream(buff, "r")

    x = fd.peek(10)
    x = fd.read()
    assert len(x) == 60


@pytest.mark.remote_data()
def test_urlopen(tree, httpserver):
    path = os.path.join(httpserver.tmpdir, "test.asdf")

    def get_write_fd():
        # cannot use context manager here because it closes the file
        return generic_io.get_file(open(path, "wb"), mode="w")

    def get_read_fd():
        return generic_io.get_file(urllib_request.urlopen(httpserver.url + "test.asdf"))

    with _roundtrip(tree, get_write_fd, get_read_fd) as ff:
        assert len(ff._blocks.blocks) == 2
        assert not isinstance(ff._blocks.blocks[0].cached_data, np.memmap)


@pytest.mark.remote_data()
def test_http_connection(tree, httpserver):
    path = os.path.join(httpserver.tmpdir, "test.asdf")

    def get_write_fd():
        # cannot use context manager here because it closes the file
        return generic_io.get_file(open(path, "wb"), mode="w")

    def get_read_fd():
        fd = generic_io.get_file(httpserver.url + "test.asdf")
        # This is to check for a "feature" in Python 3.x that reading zero
        # bytes from a socket causes it to stop.  We have code in generic_io.py
        # to workaround it.
        fd.read(0)
        return fd

    with _roundtrip(tree, get_write_fd, get_read_fd) as ff:
        assert len(ff._blocks.blocks) == 2
        assert (ff.tree["science_data"] == tree["science_data"]).all()


def test_exploded_filesystem(tree, tmp_path):
    path = os.path.join(str(tmp_path), "test.asdf")

    def get_write_fd():
        return generic_io.get_file(path, mode="w")

    def get_read_fd():
        return generic_io.get_file(path, mode="r")

    with _roundtrip(tree, get_write_fd, get_read_fd, write_options={"all_array_storage": "external"}) as ff:
        assert len(ff._blocks.blocks) == 0


def test_exploded_filesystem_fail(tree, tmp_path):
    path = os.path.join(str(tmp_path), "test.asdf")

    def get_write_fd():
        return generic_io.get_file(path, mode="w")

    def get_read_fd():
        fd = io.BytesIO()
        with open(path, mode="rb") as fd2:
            fd.write(fd2.read())
        fd.seek(0)
        return fd

    with get_write_fd() as fd:
        asdf.AsdfFile(tree).write_to(fd, all_array_storage="external")

    with get_read_fd() as fd, asdf.open(fd) as ff, pytest.raises(ValueError, match=r"Resolved to relative URL"):
        helpers.assert_tree_match(tree, ff.tree)


@pytest.mark.remote_data()
def test_exploded_http(tree, httpserver):
    path = os.path.join(httpserver.tmpdir, "test.asdf")

    def get_write_fd():
        return generic_io.get_file(path, mode="w")

    def get_read_fd():
        return generic_io.get_file(httpserver.url + "test.asdf")

    with _roundtrip(tree, get_write_fd, get_read_fd, write_options={"all_array_storage": "external"}) as ff:
        assert len(list(ff._blocks.blocks)) == 0


def test_exploded_stream_write(small_tree):
    # Writing an exploded file to an output stream should fail, since
    # we can't write "files" alongside it.

    ff = asdf.AsdfFile(small_tree)

    with pytest.raises(ValueError, match=r"Can't write external blocks, since URI of main file is unknown."):
        ff.write_to(io.BytesIO(), all_array_storage="external")


def test_exploded_stream_read(tmp_path, small_tree):
    # Reading from an exploded input file should fail, but only once
    # the data block is accessed.  This behavior is important so that
    # the tree can still be accessed even if the data is missing.

    path = os.path.join(str(tmp_path), "test.asdf")

    ff = asdf.AsdfFile(small_tree)
    ff.write_to(path, all_array_storage="external")

    with open(path, "rb") as fd:
        # This should work, so we can get the tree content
        x = generic_io.InputStream(fd, "r")
        # It's only when trying to access external data that an error occurs
        with asdf.open(x) as ff, pytest.raises(ValueError, match=r"Resolved to relative URL"):
            ff.tree["science_data"][:]


def test_unicode_open(tmp_path, small_tree):
    path = os.path.join(str(tmp_path), "test.asdf")

    ff = asdf.AsdfFile(small_tree)

    ff.write_to(path)

    with (
        open(path, encoding="utf-8") as fd,
        pytest.raises(
            ValueError,
            match=r"File-like object must be opened in binary mode.",
        ),
        asdf.open(fd),
    ):
        pass


def test_open_stdout():
    """Test ability to open/write stdout as an output stream"""
    with generic_io.get_file(sys.__stdout__, "w", close=True):
        pass


def test_invalid_obj(tmp_path):
    with pytest.raises(ValueError, match=r"Can't handle .* as a file for mode 'r'"):
        generic_io.get_file(42)

    path = os.path.join(str(tmp_path), "test.asdf")
    with (
        generic_io.get_file(path, "w") as fd,
        pytest.raises(
            ValueError,
            match=r"File is opened as 'w', but 'r' was requested",
        ),
    ):
        generic_io.get_file(fd, "r")

    url = "http://www.google.com"
    mode = "w"
    with pytest.raises(ValueError, match=f"Unable to open {url} with mode {mode}"):
        generic_io.get_file(url, mode)

    with pytest.raises(TypeError, match=r"io.StringIO objects are not supported.  Use io.BytesIO instead."):
        generic_io.get_file(io.StringIO())

    with open(path, "rb") as fd, pytest.raises(ValueError, match=r"File is opened as 'rb', but 'w' was requested"):
        generic_io.get_file(fd, "w")


def test_nonseekable_file(tmp_path):
    base = io.FileIO

    class FileWrapper(base):
        def tell(self):
            raise OSError

        def seekable(self):
            return False

        def readable(self):
            return True

        def writable(self):
            return True

    with FileWrapper(os.path.join(str(tmp_path), "test.asdf"), "wb") as fd:
        assert isinstance(generic_io.get_file(fd, "w"), generic_io.OutputStream)
        with pytest.raises(ValueError, match=r"File .* could not be opened in 'rw' mode"):
            generic_io.get_file(fd, "rw")

    with FileWrapper(os.path.join(str(tmp_path), "test.asdf"), "rb") as fd:
        assert isinstance(generic_io.get_file(fd, "r"), generic_io.InputStream)


def test_relative_uri():
    assert generic_io.relative_uri("http://www.google.com", "file://local") == "file://local"


@pytest.mark.parametrize("protocol", ["http", "asdf"])
def test_resolve_uri(protocol):
    """
    Confirm that the patched urllib.parse is handling
    asdf:// URIs correctly.
    """
    assert (
        generic_io.resolve_uri(f"{protocol}://somewhere.org/some-schema", "#/definitions/foo")
        == f"{protocol}://somewhere.org/some-schema#/definitions/foo"
    )

    assert (
        generic_io.resolve_uri(
            f"{protocol}://somewhere.org/path/to/some-schema",
            "../../some/other/path/to/some-other-schema",
        )
        == f"{protocol}://somewhere.org/some/other/path/to/some-other-schema"
    )


def test_arbitrary_file_object():
    class Wrapper:
        def __init__(self, init):
            self._fd = init

    class Random:
        def seek(self, *args):
            return self._fd.seek(*args)

        def tell(self, *args):
            return self._fd.tell(*args)

    class Reader(Wrapper):
        def read(self, *args):
            return self._fd.read(*args)

    class RandomReader(Reader, Random):
        pass

    class Writer(Wrapper):
        def write(self, *args):
            return self._fd.write(*args)

    class RandomWriter(Writer, Random):
        pass

    class All(Reader, Writer, Random):
        pass

    buff = io.BytesIO()
    assert isinstance(generic_io.get_file(Reader(buff), "r"), generic_io.InputStream)
    assert isinstance(generic_io.get_file(Writer(buff), "w"), generic_io.OutputStream)
    assert isinstance(generic_io.get_file(RandomReader(buff), "r"), generic_io.MemoryIO)
    assert isinstance(generic_io.get_file(RandomWriter(buff), "w"), generic_io.MemoryIO)
    assert isinstance(generic_io.get_file(All(buff), "rw"), generic_io.MemoryIO)
    assert isinstance(generic_io.get_file(All(buff), "r"), generic_io.MemoryIO)
    assert isinstance(generic_io.get_file(All(buff), "w"), generic_io.MemoryIO)

    with pytest.raises(ValueError, match=r"Can't handle .* as a file for mode 'w'"):
        generic_io.get_file(Reader(buff), "w")

    with pytest.raises(ValueError, match=r"Can't handle .* as a file for mode 'r'"):
        generic_io.get_file(Writer(buff), "r")


def test_check_bytes(tmp_path):
    with open(os.path.join(str(tmp_path), "test.asdf"), "w", encoding="utf-8") as fd:
        assert generic_io._check_bytes(fd, "r") is False
        assert generic_io._check_bytes(fd, "rw") is False
        assert generic_io._check_bytes(fd, "w") is False

    with open(os.path.join(str(tmp_path), "test.asdf"), "wb") as fd:
        assert generic_io._check_bytes(fd, "r") is True
        assert generic_io._check_bytes(fd, "rw") is True
        assert generic_io._check_bytes(fd, "w") is True


def test_truncated_reader():
    """
    Tests several edge cases for _TruncatedReader.read()

    Includes regression test for
    https://github.com/asdf-format/asdf/pull/181
    """

    # TODO: Should probably break this up into multiple test cases

    fd = generic_io.RandomAccessFile(io.BytesIO(), "rw")
    content = b"a" * 100 + b"b"
    fd.write(content)
    fd.seek(0)

    # Simple cases where the delimiter is not found at all
    tr = generic_io._TruncatedReader(fd, b"x", 1)
    with pytest.raises(exceptions.DelimiterNotFoundError, match=r"b'x' not found"):
        tr.read()

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"x", 1)
    assert tr.read(100) == content[:100]
    assert tr.read(1) == content[100:]
    with pytest.raises(exceptions.DelimiterNotFoundError, match=r"b'x' not found"):
        tr.read()

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"x", 1, exception=False)
    assert tr.read() == content

    # No delimiter but with 'initial_content'
    init = b"abcd"
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"x", 1, initial_content=init, exception=False)
    assert tr.read(100) == (init + content)[:100]
    assert tr.read() == (init + content)[100:]

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"x", 1, initial_content=init, exception=False)
    assert tr.read() == init + content

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"x", 1, initial_content=init, exception=False)
    assert tr.read(2) == init[:2]
    assert tr.read() == init[2:] + content

    # Some tests of a single character delimiter
    # Add some trailing data after the delimiter
    fd.seek(0, 2)
    fd.write(b"ffff")

    # Delimiter not included in read
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"b", 1)
    assert tr.read(100) == content[:100]
    assert tr.read() == b""

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"b", 1)
    assert tr.read() == content[:100]

    # Delimiter included
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"b", 1, include=True)
    assert tr.read() == content[:101]
    assert tr.read() == b""

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"b", 1, include=True)
    assert tr.read(101) == content[:101]
    assert tr.read() == b""

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"b", 1, include=True)
    assert tr.read(102) == content[:101]
    assert tr.read() == b""

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, b"b", 1, include=True)
    assert tr.read(100) == content[:100]
    assert tr.read(1) == content[100:101]
    assert tr.read() == b""

    # Longer delimiter with variable length
    content = b"a" * 100 + b"\n...\n" + b"ffffff"
    delimiter = rb"\r?\n\.\.\.((\r?\n)|$)"
    readahead = 7

    fd = generic_io.RandomAccessFile(io.BytesIO(), "rw")
    fd.write(content)

    # Delimiter not included in read
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead)
    assert tr.read() == content[:100]
    assert tr.read() == b""

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead)
    assert tr.read(100) == content[:100]
    assert tr.read() == b""

    # (read just up to the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead)
    assert tr.read(99) == content[:99]
    assert tr.read() == content[99:100]
    assert tr.read() == b""

    # (read partway into the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead)
    assert tr.read(99) == content[:99]
    assert tr.read(2) == content[99:100]
    assert tr.read() == b""

    # (read well past the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead)
    assert tr.read(99) == content[:99]
    assert tr.read(50) == content[99:100]
    assert tr.read() == b""

    # Same as the previous set of tests, but including the delimiter
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, include=True)
    assert tr.read() == content[:105]
    assert tr.read() == b""

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, include=True)
    assert tr.read(105) == content[:105]
    assert tr.read() == b""

    # (read just up to the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, include=True)
    assert tr.read(99) == content[:99]
    assert tr.read() == content[99:105]
    assert tr.read() == b""

    # (read partway into the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, include=True)
    assert tr.read(99) == content[:99]
    assert tr.read(2) == content[99:101]
    assert tr.read() == content[101:105]
    assert tr.read() == b""

    # (read well past the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, include=True)
    assert tr.read(99) == content[:99]
    assert tr.read(50) == content[99:105]
    assert tr.read() == b""

    # Same sequence of tests but with some 'initial_content'
    init = b"abcd"

    # Delimiter not included in read
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, initial_content=init)
    assert tr.read() == (init + content[:100])
    assert tr.read() == b""

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, initial_content=init)
    assert tr.read(100) == (init + content[:96])
    assert tr.read() == content[96:100]
    assert tr.read() == b""

    # (read just up to the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, initial_content=init)
    assert tr.read(99) == (init + content[:95])
    assert tr.read() == content[95:100]
    assert tr.read() == b""

    # (read partway into the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, initial_content=init)
    assert tr.read(99) == (init + content[:95])
    assert tr.read(6) == content[95:100]
    assert tr.read() == b""

    # (read well past the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, initial_content=init)
    assert tr.read(99) == (init + content[:95])
    assert tr.read(50) == content[95:100]
    assert tr.read() == b""

    # Same as the previous set of tests, but including the delimiter
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, include=True, initial_content=init)
    assert tr.read() == (init + content[:105])
    assert tr.read() == b""

    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, include=True, initial_content=init)
    assert tr.read(105) == (init + content[:101])
    assert tr.read() == content[101:105]
    assert tr.read() == b""

    # (read just up to the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, include=True, initial_content=init)
    assert tr.read(103) == (init + content[:99])
    assert tr.read() == content[99:105]
    assert tr.read() == b""

    # (read partway into the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, include=True, initial_content=init)
    assert tr.read(99) == (init + content[:95])
    assert tr.read(6) == content[95:101]
    assert tr.read() == content[101:105]
    assert tr.read() == b""

    # (read well past the delimiter)
    fd.seek(0)
    tr = generic_io._TruncatedReader(fd, delimiter, readahead, include=True, initial_content=init)
    assert tr.read(99) == (init + content[:95])
    assert tr.read(50) == content[95:105]
    assert tr.read() == b""


def test_blocksize(tree, tmp_path):
    path = os.path.join(str(tmp_path), "test.asdf")

    def get_write_fd():
        # cannot use context manager here because it closes the file
        return generic_io.get_file(open(path, "wb"), mode="w", close=True)

    def get_read_fd():
        # Must open with mode=rw in order to get memmapped data
        # cannot use context manager here because it closes the file
        return generic_io.get_file(open(path, "r+b"), mode="rw", close=True)

    with config_context() as config:
        config.io_block_size = 1233  # make sure everything works with a strange blocksize
        with _roundtrip(tree, get_write_fd, get_read_fd) as ff:
            assert ff._fd.block_size == 1233


def test_io_subclasses(tmp_path):
    ref = b"0123456789"

    b = io.BytesIO(b"0123456789")
    b.seek(0)
    f = generic_io.get_file(b)
    r = f.read(len(ref))
    assert r == ref, (r, ref)
    f.close()

    b = io.BytesIO(b"0123456789")
    b.seek(0)
    br = io.BufferedReader(io.BytesIO(ref))
    f = generic_io.get_file(b)
    assert r == ref, (r, ref)
    f.close()

    b = io.BytesIO(b"")
    bw = io.BufferedWriter(b)
    f = generic_io.get_file(bw, mode="w")
    f.write(ref)
    b.seek(0)
    r = b.read(len(ref))
    assert r == ref, (r, ref)
    f.close()

    b = io.BytesIO(b"")
    br = io.BufferedRandom(b)
    f = generic_io.get_file(br, mode="rw")
    f.write(ref)
    f.seek(0)
    r = f.read(len(ref))
    assert r == ref, (r, ref)
    f.close()


def test_fsspec(tmp_path):
    """
    Issue #1146 reported errors when opening a fsspec 'file'
    This is a regression test for the fix in PR #1226
    """
    fsspec = pytest.importorskip("fsspec")

    ref = b"01234567890"
    fn = tmp_path / "test"

    with fsspec.open(fn, mode="bw+") as f:
        f.write(ref)
        f.seek(0)
        gf = generic_io.get_file(f)
        r = gf.read(len(ref))
        assert r == ref, (r, ref)

        gf.seek(0)
        arr = gf.read_into_array(len(ref))
        for ai, i in zip(arr, ref):
            assert ai == i


@pytest.mark.remote_data()
def test_fsspec_http(httpserver):
    """
    Issue #1146 reported errors when opening a fsspec url (using the http
    filesystem)
    This is a regression test for the fix in PR #1228
    """
    fsspec = pytest.importorskip("fsspec")

    ref = b"01234567890"
    path = os.path.join(httpserver.tmpdir, "test")

    with open(path, "wb") as f:
        f.write(ref)

    fn = httpserver.url + "test"
    with fsspec.open(fn) as f:
        gf = generic_io.get_file(f)
        r = gf.read(len(ref))
        assert r == ref, (r, ref)

        gf.seek(0)
        arr = gf.read_into_array(len(ref))
        for ai, i in zip(arr, ref):
            assert ai == i


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="Cannot test write file permissions on windows",
)
@pytest.mark.parametrize("umask", range(512))
def test_write_file_permissions(tmp_path, umask):
    previous_umask = os.umask(umask)
    fn = tmp_path / "foo"
    with generic_io.get_file(fn, mode="w"):
        pass
    permissions = os.stat(fn)[stat.ST_MODE] & 0o777
    os.umask(previous_umask)
    target_permissions = generic_io._FILE_PERMISSIONS_NO_EXECUTE & ~umask
    assert permissions == target_permissions



import numpy as np
import pytest

import asdf._block.io as bio
from asdf import constants, generic_io
from asdf._block import reader, writer


@pytest.mark.parametrize("lazy", [True, False])
@pytest.mark.parametrize("index", [True, False])
@pytest.mark.parametrize("padding", [True, False, 0.1, 0.9])
@pytest.mark.parametrize("compression", [None, b"zlib"])
@pytest.mark.parametrize("stream", [True, False])
@pytest.mark.parametrize("seekable", [True, False])
def test_write_blocks(tmp_path, lazy, index, padding, compression, stream, seekable):
    data = [np.ones(10, dtype=np.uint8), np.zeros(5, dtype=np.uint8), None]
    if lazy:
        blocks = [writer.WriteBlock(lambda bd=d: bd, compression=compression) for d in data]
    else:
        blocks = [writer.WriteBlock(d, compression=compression) for d in data]
    if stream:
        streamed_block = writer.WriteBlock(np.ones(15, dtype=np.uint8))
    else:
        streamed_block = None
    fn = tmp_path / "test.bin"
    with generic_io.get_file(fn, mode="w") as fd:
        if not seekable:
            fd.seekable = lambda: False
        writer.write_blocks(fd, blocks, padding=padding, streamed_block=streamed_block, write_index=index)
    with generic_io.get_file(fn, mode="r") as fd:
        if index and not stream:
            assert bio.find_block_index(fd) is not None
        else:
            assert bio.find_block_index(fd) is None
        fd.seek(0)
        read_blocks = reader.read_blocks(fd)
        if stream:
            assert len(read_blocks) == (len(data) + 1)
        else:
            assert len(read_blocks) == len(data)
        for r, d in zip(read_blocks, data):
            if d is None:
                assert r.data.size == 0
            else:
                np.testing.assert_array_equal(r.data, d)
            if compression is not None:
                assert r.header["compression"] == compression
            if padding:
                assert r.header["allocated_size"] > r.header["used_size"]
        if stream:
            read_stream_block = read_blocks[-1]
            np.testing.assert_array_equal(read_stream_block.data, streamed_block.data)
            assert read_stream_block.header["flags"] & constants.BLOCK_FLAG_STREAMED


def _raise_illegal_seek():
    raise OSError("Illegal seek")


@pytest.mark.parametrize("stream", [True, False])
@pytest.mark.parametrize("index", [True, False])
@pytest.mark.parametrize("tell", [0, None, _raise_illegal_seek])
def test_non_seekable_files_with_odd_tells(tmp_path, stream, index, tell):
    """
    Some non-seekable files have odd 'tell' results. See:
    https://github.com/asdf-format/asdf/issues/1545
    https://github.com/asdf-format/asdf/issues/1542

    These can produce invalid block indices which should not be written
    to the ASDF file.
    """
    data = [np.ones(10, dtype=np.uint8), np.zeros(5, dtype=np.uint8), None]
    blocks = [writer.WriteBlock(d) for d in data]
    if stream:
        streamed_block = writer.WriteBlock(np.ones(15, dtype=np.uint8))
    else:
        streamed_block = None
    fn = tmp_path / "test.bin"
    with generic_io.get_file(fn, mode="w") as fd:
        fd.seekable = lambda: False
        if callable(tell):
            fd.tell = tell
        else:
            fd.tell = lambda: tell
        writer.write_blocks(fd, blocks, streamed_block=streamed_block, write_index=index)
    with generic_io.get_file(fn, mode="r") as fd:
        assert bio.find_block_index(fd) is None
        fd.seek(0)
        read_blocks = reader.read_blocks(fd)
        if stream:
            assert len(read_blocks) == (len(data) + 1)
        else:
            assert len(read_blocks) == len(data)
        for r, d in zip(read_blocks, data):
            if d is None:
                assert r.data.size == 0
            else:
                np.testing.assert_array_equal(r.data, d)
        if stream:
            read_stream_block = read_blocks[-1]
            np.testing.assert_array_equal(read_stream_block.data, streamed_block.data)
            assert read_stream_block.header["flags"] & constants.BLOCK_FLAG_STREAMED



import numpy as np

from asdf.extension import Converter


class IntegerConverter(Converter):
    tags = [
        "tag:stsci.edu:asdf/core/integer-1.0.0",
        "tag:stsci.edu:asdf/core/integer-1.1.0",
        "tag:stsci.edu:asdf/core/integer-1.2.0",
    ]
    types = ["asdf.tags.core.integer.IntegerType"]

    def to_yaml_tree(self, obj, tag, ctx):
        abs_value = int(np.abs(obj._value))

        # pack integer value into 32-bit words
        words = []
        value = abs_value
        while value > 0:
            words.append(value & 0xFFFFFFFF)
            value >>= 32

        array = np.array(words, dtype=np.uint32)

        tree = {}
        ctx.set_array_storage(array, obj._storage)
        tree["words"] = array
        tree["sign"] = obj._sign
        tree["string"] = str(int(obj._value))

        return tree

    def from_yaml_tree(self, node, tag, ctx):
        from asdf.tags.core.integer import IntegerType

        value = 0
        for x in node["words"][::-1]:
            value <<= 32
            value |= int(x)

        if node["sign"] == "-":
            value = -value

        return IntegerType(value)



import os
import re
from contextlib import contextmanager

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import asdf
from asdf import constants
from asdf._block import io as bio
from asdf._commands import main

RNG = np.random.default_rng(42)


@pytest.fixture(params=asdf.versioning.supported_versions)
def version(request):
    return request.param


@pytest.fixture()
def create_editor(tmp_path):
    """
    Fixture providing a function that generates an editor script.
    """

    def _create_editor(pattern, replacement):
        if isinstance(pattern, str):
            pattern = pattern.encode("utf-8")
        if isinstance(replacement, str):
            replacement = replacement.encode("utf-8")

        editor_path = tmp_path / "editor.py"

        content = f"""import re
import sys

with open(sys.argv[1], "rb") as file:
    content = file.read()

content = re.sub({pattern!r}, {replacement!r}, content, flags=(re.DOTALL | re.MULTILINE))

with open(sys.argv[1], "wb") as file:
    file.write(content)
"""

        with editor_path.open("w") as file:
            file.write(content)

        return f"python3 {editor_path}"

    return _create_editor


@contextmanager
def file_not_modified(path):
    """
    Assert that a file was not modified during the context.
    """
    original_mtime = os.stat(path).st_mtime_ns

    yield

    assert os.stat(path).st_mtime_ns == original_mtime


@pytest.fixture()
def mock_input(monkeypatch):
    """
    Fixture providing a function that mocks the edit module's
    built-in input function.
    """

    @contextmanager
    def _mock_input(pattern, response):
        called = False

        def _input(prompt=None):
            nonlocal called
            called = True
            assert prompt is not None
            assert re.match(pattern, prompt)
            return response

        with monkeypatch.context() as m:
            m.setattr("builtins.input", _input)
            yield

        assert called, "input was not called as expected"

    return _mock_input


@pytest.fixture(autouse=True)
def _default_mock_input(monkeypatch):
    """
    Fixture that raises an error when the program
    requests unexpected input.
    """

    def _input(prompt=None):
        msg = f"Received unexpected request for input: {prompt}"
        raise AssertionError(msg)

    monkeypatch.setattr("builtins.input", _input)


def test_no_blocks(tmp_path, create_editor, version):
    file_path = str(tmp_path / "test.asdf")

    with asdf.AsdfFile(version=version) as af:
        af["foo"] = "bar"
        af.write_to(file_path)

    os.environ["EDITOR"] = create_editor(r"foo: bar", "foo: baz")

    assert main.main_from_args(["edit", file_path]) == 0

    with asdf.open(file_path) as af:
        assert af["foo"] == "baz"


def test_no_blocks_increase_size(tmp_path, create_editor, version):
    file_path = str(tmp_path / "test.asdf")

    with asdf.AsdfFile(version=version) as af:
        af["foo"] = "bar"
        af.write_to(file_path)

    new_value = "a" * 32768
    os.environ["EDITOR"] = create_editor(r"foo: bar", f"foo: {new_value}")

    # With no blocks, we can expand the existing file, so this case
    # shouldn't require confirmation from the user.
    assert main.main_from_args(["edit", file_path]) == 0

    with asdf.open(file_path) as af:
        assert af["foo"] == new_value


def test_no_blocks_decrease_size(tmp_path, create_editor, version):
    file_path = str(tmp_path / "test.asdf")

    original_value = "a" * 32768

    with asdf.AsdfFile(version=version) as af:
        af["foo"] = original_value
        af.write_to(file_path)

    os.environ["EDITOR"] = create_editor(f"foo: {original_value}", "foo: bar")

    assert main.main_from_args(["edit", file_path]) == 0

    with asdf.open(file_path) as af:
        assert af["foo"] == "bar"


def confirm_valid_block_index(file_path):
    # make sure the block index is valid
    with asdf.generic_io.get_file(file_path, "r") as f:
        block_index_offset = bio.find_block_index(f)
        assert block_index_offset is not None
        block_index = bio.read_block_index(f, block_index_offset)
        for block_offset in block_index:
            f.seek(block_offset)
            assert f.read(len(constants.BLOCK_MAGIC)) == constants.BLOCK_MAGIC


def test_with_blocks(tmp_path, create_editor, version):
    file_path = str(tmp_path / "test.asdf")

    array1 = RNG.normal(size=100)
    array2 = RNG.normal(size=100)
    with asdf.AsdfFile(version=version) as af:
        af["array1"] = array1
        af["array2"] = array2
        af["foo"] = "bar"
        af.write_to(file_path)

    os.environ["EDITOR"] = create_editor(r"foo: bar", "foo: baz")

    assert main.main_from_args(["edit", file_path]) == 0

    with asdf.open(file_path) as af:
        assert af["foo"] == "baz"
        assert_array_equal(af["array1"], array1)
        assert_array_equal(af["array2"], array2)

    confirm_valid_block_index(file_path)


def test_with_blocks_increase_size(tmp_path, create_editor, version, mock_input):
    file_path = str(tmp_path / "test.asdf")

    array1 = RNG.normal(size=100)
    array2 = RNG.normal(size=100)
    with asdf.AsdfFile(version=version) as af:
        af["array1"] = array1
        af["array2"] = array2
        af["foo"] = "bar"
        af.write_to(file_path)

    new_value = "a" * 32768
    os.environ["EDITOR"] = create_editor(r"foo: bar", f"foo: {new_value}")

    # Abort without updating the file
    with mock_input(r"\(c\)ontinue or \(a\)bort\?", "a"), file_not_modified(file_path):
        assert main.main_from_args(["edit", file_path]) == 1

    # Agree to allow the file to be rewritten
    with mock_input(r"\(c\)ontinue or \(a\)bort\?", "c"):
        assert main.main_from_args(["edit", file_path]) == 0

    with asdf.open(file_path) as af:
        assert af["foo"] == new_value
        assert_array_equal(af["array1"], array1)
        assert_array_equal(af["array2"], array2)

    confirm_valid_block_index(file_path)


def test_with_blocks_decrease_size(tmp_path, create_editor, version):
    file_path = str(tmp_path / "test.asdf")

    original_value = "a" * 32768

    array1 = RNG.normal(size=100)
    array2 = RNG.normal(size=100)
    with asdf.AsdfFile(version=version) as af:
        af["array1"] = array1
        af["array2"] = array2
        af["foo"] = original_value
        af.write_to(file_path)

    os.environ["EDITOR"] = create_editor(f"foo: {original_value}", "foo: bar")

    assert main.main_from_args(["edit", file_path]) == 0

    with asdf.open(file_path) as af:
        assert af["foo"] == "bar"
        assert_array_equal(af["array1"], array1)
        assert_array_equal(af["array2"], array2)

    confirm_valid_block_index(file_path)


def test_no_changes(tmp_path, create_editor, version):
    file_path = str(tmp_path / "test.asdf")

    with asdf.AsdfFile(version=version) as af:
        af["foo"] = "bar"
        af.write_to(file_path)

    os.environ["EDITOR"] = create_editor(r"non-existent-string", "non-existent-string")

    with file_not_modified(file_path):
        assert main.main_from_args(["edit", file_path]) == 0


def test_update_asdf_standard_version(tmp_path, create_editor, version, mock_input):
    file_path = str(tmp_path / "test.asdf")

    with asdf.AsdfFile(version=version) as af:
        af["foo"] = "bar"
        af.write_to(file_path)

    os.environ["EDITOR"] = create_editor(r"^#ASDF_STANDARD .*?$", "#ASDF_STANDARD 999.999.999")

    with file_not_modified(file_path), mock_input(r"\(c\)ontinue editing or \(a\)bort\?", "a"):
        assert main.main_from_args(["edit", file_path]) == 1


def test_update_yaml_version(tmp_path, create_editor, version, mock_input):
    file_path = str(tmp_path / "test.asdf")

    with asdf.AsdfFile(version=version) as af:
        af["foo"] = "bar"
        af.write_to(file_path)

    os.environ["EDITOR"] = create_editor(r"^%YAML 1.1$", "%YAML 1.2")

    with file_not_modified(file_path), mock_input(r"\(c\)ontinue editing or \(a\)bort\?", "a"):
        assert main.main_from_args(["edit", file_path]) == 1


def test_bad_yaml(tmp_path, create_editor, version, mock_input):
    file_path = str(tmp_path / "test.asdf")

    with asdf.AsdfFile(version=version) as af:
        af["foo"] = "bar"
        af.write_to(file_path)

    os.environ["EDITOR"] = create_editor(r"foo: bar", "foo: [")

    with file_not_modified(file_path), mock_input(r"\(c\)ontinue editing or \(a\)bort\?", "a"):
        assert main.main_from_args(["edit", file_path]) == 1


def test_validation_failure(tmp_path, create_editor, version, mock_input):
    file_path = str(tmp_path / "test.asdf")

    with asdf.AsdfFile(version=version) as af:
        af["array"] = np.arange(100)
        af.write_to(file_path)

    os.environ["EDITOR"] = create_editor(r"byteorder: .*?$", "byteorder: med")

    with file_not_modified(file_path), mock_input(r"\(c\)ontinue editing, \(f\)orce update, or \(a\)bort\?", "a"):
        assert main.main_from_args(["edit", file_path]) == 1

    with mock_input(r"\(c\)ontinue editing, \(f\)orce update, or \(a\)bort\?", "f"):
        assert main.main_from_args(["edit", file_path]) == 0

    with open(file_path, "rb") as f:
        content = f.read()
        assert b"byteorder: med" in content


def test_asdf_open_failure(tmp_path, create_editor, version, mock_input):
    file_path = str(tmp_path / "test.asdf")

    with asdf.AsdfFile(version=version) as af:
        af["foo"] = "bar"
        af.write_to(file_path)

    os.environ["EDITOR"] = create_editor(r"^#ASDF .*?$", "#HJKL 1.0.0")

    with file_not_modified(file_path), mock_input(r"\(c\)ontinue editing or \(a\)bort\?", "a"):
        assert main.main_from_args(["edit", file_path]) == 1


def test_non_asdf_file(tmp_path):
    file_path = str(tmp_path / "test.asdf")

    with open(file_path, "w") as f:
        f.write("Dear diary...")

    with file_not_modified(file_path):
        assert main.main_from_args(["edit", file_path]) == 1



import pytest


@pytest.fixture(scope="session")
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setenv("AWS_ACCESS_KEY_ID", "testing")
        monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "testing")
        monkeypatch.setenv("AWS_SECURITY_TOKEN", "testing")
        monkeypatch.setenv("AWS_SESSION_TOKEN", "testing")
        monkeypatch.setenv("AWS_DEFAULT_REGION", "us-east-1")
        yield


@pytest.fixture(scope="session")
def s3(aws_credentials):
    """
    Return a mocked S3 client
    """
    IP = "127.0.0.1"
    PORT = 3000
    ENDPOINT_URL = f"http://{IP}:{PORT}"
    boto3 = pytest.importorskip("boto3")
    moto_server = pytest.importorskip("moto.server")
    server = moto_server.ThreadedMotoServer(ip_address=IP, port=PORT)
    server.start()
    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setenv("FSSPEC_S3_ENDPOINT_URL", ENDPOINT_URL)
        yield boto3.client("s3", endpoint_url=ENDPOINT_URL)



from numbers import Integral


class IntegerType:
    """
    Enables the storage of arbitrarily large integer values

    The ASDF specification mandates that integer literals in the tree can be no
    larger than 64 bits. Use of this class enables the storage of arbitrarily
    large integer values.

    When reading files that contain arbitrarily large integers, the values that
    are restored in the tree will be raw Python `int` instances.

    Parameters
    ----------

    value: `numbers.Integral`
        A Python integral value (e.g. `int` or `numpy.integer`)

    storage_type: `str`, optional
        Optionally overrides the storage type of the array used to represent
        the integer value. Valid values are "internal" (the default) and
        "inline"

    Examples
    --------

    >>> import asdf
    >>> import random
    >>> # Create a large integer value
    >>> largeval = random.getrandbits(100)
    >>> # Store the large integer value to the tree using asdf.IntegerType
    >>> tree = dict(largeval=asdf.IntegerType(largeval))
    >>> with asdf.AsdfFile(tree) as af:
    ...     af.write_to('largeval.asdf')
    >>> with asdf.open('largeval.asdf') as aa:
    ...     assert aa['largeval'] == largeval
    """

    def __init__(self, value, storage_type="internal"):
        if storage_type not in ["internal", "inline"]:
            msg = f"storage_type '{storage_type}' is not a recognized storage type"
            raise ValueError(msg)
        self._value = value
        self._sign = "-" if value < 0 else "+"
        self._storage = storage_type

    def __int__(self):
        return int(self._value)

    def __float__(self):
        return float(self._value)

    def __eq__(self, other):
        if isinstance(other, Integral):
            return self._value == other

        if isinstance(other, IntegerType):
            return self._value == other._value

        msg = f"Can't compare IntegralType to unknown type: {type(other)}"
        raise ValueError(msg)

    def __repr__(self):
        return f"IntegerType({self._value})"



from asdf.resource import JsonschemaResourceMapping


def get_extensions():
    """
    Get the extension instances for the core extensions.  This method is registered with the
    asdf.extensions entry point.

    Returns
    -------
    list of asdf.extension.Extension
    """
    from . import _extensions

    return _extensions.EXTENSIONS


def get_json_schema_resource_mappings():
    return [
        JsonschemaResourceMapping(),
    ]



"""
asdf: Python library for reading and writing Advanced Scientific
Data Format (ASDF) files
"""

__all__ = [
    "AsdfFile",
    "ExternalArrayReference",
    "IntegerType",
    "Stream",
    "ValidationError",
    "__version__",
    "config_context",
    "dump",
    "dumps",
    "get_config",
    "info",
    "load",
    "loads",
    "open",
]


from ._asdf import AsdfFile
from ._asdf import open_asdf as open
from ._convenience import info
from ._dump import dump, dumps, load, loads
from ._version import version as __version__
from .config import config_context, get_config
from .exceptions import ValidationError
from .tags.core import IntegerType, Stream
from .tags.core.external_reference import ExternalArrayReference



class TagDefinition:
    """
    Container for properties of a custom YAML tag.

    Parameters
    ----------
    tag_uri : str
        Tag URI.
    schema_uris : str, optional
        URI of the schema that should be used to validate objects
        with this tag.
    title : str, optional
        Short description of the tag.
    description : str, optional
        Long description of the tag.
    """

    def __init__(self, tag_uri, *, schema_uris=None, title=None, description=None):
        if "*" in tag_uri:
            msg = "URI patterns are not permitted in TagDefinition"
            raise ValueError(msg)

        self._tag_uri = tag_uri

        if schema_uris is None:
            self._schema_uris = []

        elif isinstance(schema_uris, list):
            self._schema_uris = schema_uris

        else:
            self._schema_uris = [schema_uris]

        self._title = title
        self._description = description

    @property
    def tag_uri(self):
        """
        Get the tag URI.

        Returns
        -------
        str
        """
        return self._tag_uri

    @property
    def schema_uris(self):
        """
        Get the URIs of the schemas that should be used to validate
        objects with this tag.

        Returns
        -------
        list
        """
        return self._schema_uris

    @property
    def title(self):
        """
        Get the short description of the tag.

        Returns
        -------
        str or None
        """
        return self._title

    @property
    def description(self):
        """
        Get the long description of the tag.

        Returns
        -------
        str or None
        """
        return self._description

    def __repr__(self):
        return f"<TagDefinition URI: {self.tag_uri}>"
